##############################################################################
#
#                        Crossbar.io
#     Copyright (C) Crossbar.io Technologies GmbH. All rights reserved.
#
##############################################################################

import sys
import os
import json
import pkg_resources
from collections import OrderedDict

import click
import psutil

import txaio
from txaio import make_logger

from twisted.internet.defer import Deferred, inlineCallbacks
from twisted.internet.ssl import optionsForClientTLS
from twisted.internet.task import LoopingCall

from autobahn import wamp
from autobahn.twisted.util import sleep
from autobahn.twisted.wamp import ApplicationRunner
from autobahn.wamp.types import ComponentConfig, CallOptions
from autobahn.wamp.exception import ApplicationError
from autobahn.websocket.util import parse_url

from crossbar._util import hl, hltype, hlid
from crossbar.common import checkconfig
from crossbar.node import node, controller

from crossbar.edge.node.management import NodeManagementBridgeSession, NodeManagementSession

try:
    from crossbar.edge.node.docker import DockerClient
    _HAS_DOCKER = True
except ImportError:
    _HAS_DOCKER = False

_ALLOWED_ACTIVATION_TAGS = [
    'created-at', 'management-url', 'management-realm', 'management-realm-oid', 'node-oid', 'node-authid',
    'node-cluster-ip', 'activation-code', 'public-key-ed25519'
]


def _parse_activation_file(path):
    """
    Internal helper. This parses a ``key.activation`` file and returns a dict mapping tags -> values.

    .. code-block::console

        Crossbar.io node activation

        created-at: 2020-07-05T11:49:59.125Z
        management-url: ws://localhost:9000/ws
        management-realm: default
        management-realm-oid: 6e8117fb-5bd8-4e83-860c-decefa1e95ac
        node-oid: 664e99a6-6a65-4f64-a95e-46ac9c28c80e
        node-authid: node-664e99
        activation-code: P33W-GS4H-5L4Q
        public-key-ed25519: 22c6e16005dfb0824466e35ae4b4f71746230628c2dec233f3b8cba22c4acce8
    """
    if not os.path.exists(path):
        raise Exception('activation file path "{}" does not exist'.format(path))

    if not os.path.isfile(path):
        raise Exception('activation file path "{}" exists, but is not a file'.format(path))

    tags = OrderedDict()
    with open(path, 'r') as key_file:
        got_blankline = False
        for line in key_file.readlines():
            if line.strip() == '':
                got_blankline = True
            elif got_blankline:
                tag, value = line.split(':', 1)
                tag = tag.strip().lower()
                value = value.strip()
                if tag not in _ALLOWED_ACTIVATION_TAGS:
                    raise Exception('invalid tag "{}" in activation file "{}"'.format(tag, path))
                if tag in tags:
                    raise Exception('duplicate tag "{}" in activation file "{}"'.format(tag, path))
                tags[tag] = value
    return tags


class FabricNodeControllerSession(controller.NodeController):
    """
    This is the central node controller for CF nodes.

    It derives of the node controller base class in CB and adds
    the following functionality exposed to CFC:

    - can manage a host Docker daemon
    """
    # yapf: disable
    log = make_logger()

    def onUserError(self, fail, msg):
        """
        Implements :func:`autobahn.wamp.interfaces.ISession.onUserError`
        """
        if isinstance(fail.value, ApplicationError):
            self.log.debug('{klass}.onUserError(): "{msg}"',
                           klass=self.__class__.__name__,
                           msg=fail.value.error_message())
        else:
            self.log.error(
                '{klass}.onUserError(): "{msg}"\n{traceback}',
                klass=self.__class__.__name__,
                msg=msg,
                traceback=txaio.failure_format_traceback(fail),
            )

    def __init__(self, node):
        controller.NodeController.__init__(self, node)

        # Docker client used for exposing the host Docker
        # daemon via the node management API.
        self._docker_client = None

    def onConnect(self):
        # attach to host docker daemon
        if _HAS_DOCKER:
            if self._node._enable_docker:
                if os.path.exists('/run/docker.sock'):
                    self._docker_client = DockerClient(self._node._reactor, self)
                    self._docker_client.startup()
                else:
                    self.log.warn('Docker daemon integration enabled, but Docker Unix domain socket path cannot be accessed!')
            else:
                self.log.info('Docker daemon integration disabled!')
        else:
            self.log.info('Docker unavailable or unsupported!')

        controller.NodeController.onConnect(self)

    @inlineCallbacks
    def _shutdown(self, restart=False, mode=None):
        # override base class method (without calling the base method) ..

        self.log.info('{klass}._shutdown(restart={restart}, mode={mode})',
                      klass=self.__class__.__name__, restart=restart, mode=mode)

        if self._node._manager_runner:
            self.log.warn('Stopping management uplink ..')
            yield self._node._manager_runner.stop()
            self._node._manager = None
            self._node._manager_runner = None

        if self._docker_client:
            yield self._docker_client.shutdown()
            self._docker_client = None

    @wamp.register(None)
    def get_status(self, details=None):
        """
        Return basic information about this node.

        :returns: Information on the Crossbar.io node.
        :rtype: dict
        """
        status = super(FabricNodeControllerSession, self).get_status(details=details)
        status.update({
            # the following come from CFC (and are only filled
            # when the node personality is FABRIC!)
            'management_realm':
            self._node._management_realm,
            'management_node_id':
            self._node._node_id,
            'management_session_id':
            self._node._manager._session_id if self._node._manager else None,
            'management_node_extra':
            self._node._node_extra,

            # True if remote Docker management is available
            'has_docker':
            self._docker_client is not None
        })
        return status

    #
    # Docker support
    # https://github.com/crossbario/crossbar-fabric-public/blob/master/docs/api/Management-API.md#docker-control
    #
    def _ensure_docker(self):
        if not self._docker_client:
            raise ApplicationError("crossbar.error.feature_unavailable",
                                   "Docker not available or Docker daemon integration not enabled")

    @wamp.register(None)
    @inlineCallbacks
    def get_docker_info(self, details=None):
        """
        https://github.com/crossbario/crossbar-fabric-public/blob/master/docs/api/Management-API.md#crossbarfabriccenterremotedockerget_info
        """
        self._ensure_docker()
        return (yield self._docker_client.get_info())

    @wamp.register(None)
    @inlineCallbacks
    def get_docker_ping(self, details=None):
        """
        https://github.com/crossbario/crossbar-fabric-public/blob/master/docs/api/Management-API.md#crossbarfabriccenterremotedockerget_ping
        """
        self._ensure_docker()
        return (yield self._docker_client.ping())

    @wamp.register(None)
    @inlineCallbacks
    def get_docker_version(self, details=None):
        """
        https://github.com/crossbario/crossbar-fabric-public/blob/master/docs/api/Management-API.md#crossbarfabriccenterremotedockerget_version
        """
        self._ensure_docker()
        return (yield self._docker_client.version())

    @wamp.register(None)
    @inlineCallbacks
    def get_docker_df(self, details=None):
        """
        https://github.com/crossbario/crossbar-fabric-public/blob/master/docs/api/Management-API.md#crossbarfabriccenterremotedockerget_df
        """
        self._ensure_docker()
        return (yield self._docker_client.df())

    @wamp.register(None)
    @inlineCallbacks
    def get_docker_containers(self, details=None):
        """
        https://github.com/crossbario/crossbar-fabric-public/blob/master/docs/api/Management-API.md#crossbarfabriccenterremotedockerget_containers
        """
        self._ensure_docker()
        return (yield self._docker_client.get_containers())

    @wamp.register(None)
    @inlineCallbacks
    def get_docker_container(self, container_id, details=None):
        """
        https://github.com/crossbario/crossbar-fabric-public/blob/master/docs/api/Management-API.md#crossbarfabriccenterremotedockerget_container
        """
        self._ensure_docker()
        return (yield self._docker_client.get_container(container_id))

    @wamp.register(None)
    @inlineCallbacks
    def start_docker_container(self, container_id, details=None):
        """
        https://github.com/crossbario/crossbar-fabric-public/blob/master/docs/api/Management-API.md#crossbarfabriccenterremotedockerstart_container
        """
        self._ensure_docker()
        return (yield self._docker_client.start(container_id))

    @wamp.register(None)
    @inlineCallbacks
    def create_docker_container(self, image, config={}, details=None):
        """
        https://github.com/crossbario/crossbar-fabric-public/blob/master/docs/api/Management-API.md#crossbarfabriccenterremotedockercreate_container
        """
        self._ensure_docker()
        return (yield self._docker_client.create(image, config))

    @wamp.register(None)
    @inlineCallbacks
    def stop_docker_container(self, container_id, details=None):
        """
        https://github.com/crossbario/crossbar-fabric-public/blob/master/docs/api/Management-API.md#crossbarfabriccenterremotedockerstop_container
        """
        self._ensure_docker()
        return (yield self._docker_client.container(container_id, 'stop'))

    @wamp.register(None)
    @inlineCallbacks
    def restart_docker_container(self, container_id, details=None):
        """
        https://github.com/crossbario/crossbar-fabric-public/blob/master/docs/api/Management-API.md#crossbarfabriccenterremotedockerrestart_container
        """
        self._ensure_docker()
        return (yield self._docker_client.restart(container_id))

    @wamp.register(None)
    @inlineCallbacks
    def destroy_docker_container(self, container_id, details=None):
        """
        https://github.com/crossbario/crossbar-fabric-public/blob/master/docs/api/Management-API.md#crossbarfabriccenterremotedockerremove_container
        """
        self._ensure_docker()
        return (yield self._docker_client.container(container_id, 'remove'))

    @wamp.register(None)
    @inlineCallbacks
    def pause_docker_container(self, container_id, details=None):
        """
        https://github.com/crossbario/crossbar-fabric-public/blob/master/docs/api/Management-API.md#crossbarfabriccenterremotedockerpause_container
        """
        self._ensure_docker()
        return (yield self._docker_client.container(container_id, 'pause'))

    @wamp.register(None)
    @inlineCallbacks
    def unpause_docker_container(self, container_id, details=None):
        """
        https://github.com/crossbario/crossbar-fabric-public/blob/master/docs/api/Management-API.md#crossbarfabriccenterremotedockerunpause_container
        """
        self._ensure_docker()
        return (yield self._docker_client.container(container_id, 'unpause'))

    @wamp.register(None)
    def request_docker_tty(self, id, details=None):
        """
        https://github.com/crossbario/crossbar-fabric-public/blob/master/docs/api/Management-API.md#crossbarfabriccenterremotedockerrequest_tty`
        """
        self._ensure_docker()
        return self._docker_client.request_tty(id)

    @wamp.register(None)
    @inlineCallbacks
    def watch_docker_container(self, container_id, tty_id, details=None):
        """
        https://github.com/crossbario/crossbar-fabric-public/blob/master/docs/api/Management-API.md#crossbarfabriccenterremotedockerwatch_container
        """
        self._ensure_docker()
        return (yield self._docker_client.watch(container_id, tty_id))

    @wamp.register(None)
    @inlineCallbacks
    def shell_docker_container(self, container_id, tty_id, details=None):
        """
        https://github.com/crossbario/crossbar-fabric-public/blob/master/docs/api/Management-API.md#crossbarfabriccenterremotedockershell_container
        """
        self._ensure_docker()
        return (yield self._docker_client.shell(container_id, tty_id))

    @wamp.register(None)
    @inlineCallbacks
    def backlog_docker_container(self, container_id, details=None):
        """
        https://github.com/crossbario/crossbar-fabric-public/blob/master/docs/api/Management-API.md#crossbarfabriccenterremotedockerbacklog_container
        """
        self._ensure_docker()
        return (yield self._docker_client.backlog(container_id))

    @wamp.register(None)
    @inlineCallbacks
    def keystroke_docker_container(self, container_id, data, details=None):
        """
        https://github.com/crossbario/crossbar-fabric-public/blob/master/docs/api/Management-API.md#crossbarfabriccenterremotedockerkeystroke_container
        """
        self._ensure_docker()
        return (yield self._docker_client.keystroke(container_id, data))

    @wamp.register(None)
    @inlineCallbacks
    def get_docker_images(self, details=None):
        """
        https://github.com/crossbario/crossbar-fabric-public/blob/master/docs/api/Management-API.md#crossbarfabriccenterremotedockerget_images
        """
        self._ensure_docker()
        return (yield self._docker_client.get_images())

    @wamp.register(None)
    @inlineCallbacks
    def delete_docker_image(self, image_id, details=None):
        """
        https://github.com/crossbario/crossbar-fabric-public/blob/master/docs/api/Management-API.md#crossbarfabriccenterremotedockerdelete_image
        """
        self._ensure_docker()
        return (yield self._docker_client.delete_image(image_id))

    @wamp.register(None)
    @inlineCallbacks
    def get_docker_image(self, image_id, details=None):
        """
        https://github.com/crossbario/crossbar-fabric-public/blob/master/docs/api/Management-API.md#crossbarfabriccenterremotedockerget_image
        """
        self._ensure_docker()
        return (yield self._docker_client.get_image(image_id))

    @wamp.register(None)
    @inlineCallbacks
    def remove_docker_image(self, image_id, details=None):
        """
        https://github.com/crossbario/crossbar-fabric-public/blob/master/docs/api/Management-API.md#crossbarfabriccenterremotedockerremove_image
        """
        self._ensure_docker()
        return (yield self._docker_client.image(image_id, 'remove'))

    @wamp.register(None)
    @inlineCallbacks
    def prune_docker_images(self, filter, details=None):
        """
        https://github.com/crossbario/crossbar-fabric-public/blob/master/docs/api/Management-API.md#crossbarfabriccenterremotedockerprune_image
        """
        self._ensure_docker()
        return (yield self._docker_client.prune(filter))

    @wamp.register(None)
    @inlineCallbacks
    def fs_docker_open(self, id, path, details=None):
        """
        https://github.com/crossbario/crossbar-fabric-public/blob/master/docs/api/Management-API.md#crossbarfabriccenterremotedockerfsopen
        """
        self._ensure_docker()
        return (yield self._docker_client.fs_open(id, path))

    @wamp.register(None)
    @inlineCallbacks
    def fs_docker_get(self, id, path, details=None):
        """
        https://github.com/crossbario/crossbar-fabric-public/blob/master/docs/api/Management-API.md#crossbarfabriccenterremotedockerfsget
        """
        self._ensure_docker()
        return (yield self._docker_client.fs_get(id, path))

    @wamp.register(None)
    @inlineCallbacks
    def fs_docker_put(self, id, path, data, details=None):
        """
        https://github.com/crossbario/crossbar-fabric-public/blob/master/docs/api/Management-API.md#crossbarfabriccenterremotedockerfsput
        """
        self._ensure_docker()
        return (yield self._docker_client.fs_put(id, path, data))


def compute_mgmt_uplink_config(log, cbdir, config, fn_reboot=None, use_activation_file=True, use_default_fabric=False):
    """
    Determine the transport configuration of the management uplink for this node
    in the following order (using the first one that succeeds):

    * node activation file
    * management URL environment variable
    * node configuration file
    * built-in default ("master.xbr.network")

    :param cbdir:
    :param config:
    :param fn_reboot:
    :param use_activation_file:
    :param use_default_fabric:
    :return:
    """
    fabric_transport_config = None

    # [1] pick up auto-activation files dropped by a master node (`${CBDIR}/key.activate`)
    if not fabric_transport_config and use_activation_file:
        def do_check_for_activation_file(activation_file, reboot_on_discover):
            if os.path.isfile(activation_file):
                tags = _parse_activation_file(activation_file)
                is_secure, hostname, port, _, _, _ = parse_url(tags['management-url'])
                config = {
                    'type': 'websocket',
                    'url': tags['management-url'],
                    'endpoint': {
                        'type': 'tcp',
                        'host': hostname,
                        'port': port,
                        'timeout': 5
                    }
                }
                if is_secure:
                    config['endpoint']['tls'] = {
                        'hostname': hostname
                    }
                _msg = 'Found auto-activation file "{}", using management URL "{}" - [1]'.format(activation_file,
                                                                                                 config['url'])
                log.info(click.style(_msg, fg='red', bold=True))

                if reboot_on_discover and fn_reboot:
                    # stop the node and enforce complete reboot - which will then pick up the new configuration
                    fn_reboot()

                # return the management uplink transport configuration, as derived from the activation file
                return config

        # an activation file must be placed next to the node key pair (key.pub, key.priv)
        activation_file = os.path.join(cbdir, 'key.activate')

        # check and maybe load config from activation file
        fabric_transport_config = do_check_for_activation_file(activation_file, reboot_on_discover=False)

        # if there wasn't an activation file, periodically check for ..
        if not fabric_transport_config:
            lc = LoopingCall(do_check_for_activation_file, activation_file, reboot_on_discover=True)
            lc.start(1)
            log.info('Looping call to check for node activation file started! - [1]')

    # [2] management uplink configured via env var
    if not fabric_transport_config:
        url = os.environ['CROSSBAR_FABRIC_URL'].strip() if 'CROSSBAR_FABRIC_URL' in os.environ else ''
        if url != '':
            secure, host, tcp_port, _, _, _ = parse_url(url)
            fabric_transport_config = {
                'type': 'websocket',
                'url': url,
                'endpoint': {
                    'type': 'tcp',
                    'host': host,
                    'port': tcp_port,
                    'timeout': 5
                }
            }
            if secure:
                fabric_transport_config['endpoint']['tls'] = {
                    'hostname': host
                }
            log.info(
                click.style('Using management uplink at "{}" (from envvar) - [2]'.format(url),
                            fg='red', bold=True))

    # [3] user has configured a custom management uplink in the node configuration
    if not fabric_transport_config:
        if 'controller' in config and 'fabric' in config['controller'] and config['controller']['fabric']:

            fabric_config = config['controller']['fabric']

            # allow to deactivate the management uplink connecting transport by setting "transport" to null
            fabric_transport_config = fabric_config.get('transport', None)

            if fabric_transport_config:
                url = fabric_transport_config.get('url', None)
                log.info(
                    click.style('Using management uplink at "{}" (from node configuration) - [3]'.format(url),
                                fg='red', bold=True))
            else:
                log.info(
                    click.style('Management uplink deactivated - [3]',
                                fg='red', bold=True))

    # [4] use hard-coded default management uplink
    if not fabric_transport_config and use_default_fabric:
        # default CFC (= our cloud hosted CFC service)
        fabric_transport_config = {
            'type': 'websocket',
            'url': 'wss://master.xbr.network/ws',
            'endpoint': {
                'type': 'tcp',
                'host': 'master.xbr.network',
                'port': 443,
                'timeout': 5,
                'tls': {
                    'hostname': 'master.xbr.network'
                }
            }
        }
        log.info(
            click.style(
                'Using default fabric controller at URL "{}" (from envvar) - [4]'.format(fabric_transport_config['url']),
                fg='red', bold=True))

    return fabric_transport_config


class FabricNode(node.Node):
    """
    Crossbar.io node personality.
    """
    DEFAULT_CONFIG_PATH = 'edge/node/config/bare.json'
    NODE_CONTROLLER = FabricNodeControllerSession

    def __init__(self, personality, cbdir=None, reactor=None, native_workers=None, options=None):
        node.Node.__init__(self, personality, cbdir, reactor, native_workers, options)

        # looping call that runs the node controller watchdog
        self._watchdog_looper = None

        # the node controller realm (where eg worker controller live). we explicitly use the
        # same realm as Crossbar.io OSS
        self._realm = 'crossbar'

        # enable docker daemon integration
        self._enable_docker = None

        # when running in managed mode, this will hold the bridge session
        # attached to the local management router
        self._bridge_session = None

        # if this node has a proper management uplink configured to connect to
        self._has_management_config = False

        # if this node was connected to its configured management uplink successfully at least once
        # during run-time (since last reboot) of this node
        self._was_management_connected = False

        # when we periodically check for a node activation file, the looping call for doing
        # the check - and automatically shutdown when an activation file was found (after boot)
        self._check_for_activation_file = None

        # when self._was_management_connected, the URL we've been connected to
        self._management_url = None

        # when running in managed mode, this will hold the management uplink session to
        # the crossbar master node
        self._manager = None
        self._manager_runner = None

        # the node's management realm when running in managed mode (this comes from CFC!)
        self._management_realm = None

        # the node's ID when running in managed mode (this comes from CFC!)
        self._node_id = None

        # node extra when running in managed mode (this comes from CFC!)
        self._node_extra = None

        # when the node starts up, it will connect to CFC, and then apply the
        # local node configuration, and this flag will be set. when the CFC connection
        # is lost, and then reestablished, the local node configuration should NOT
        # be applied a second time though - hence this flag
        self._local_config_applied = False

        # We really only need to see this once (?)
        self._displayed_pairing_message = False

        # for automatic ID assignment of "makers" within workers of type "xbrmm"
        self._maker_no = 1

    def load_config(self, configfile=None):
        """
        Check and load the node configuration from:

        * from ``.crossbar/config.json`` or
        * from built-in (empty) default configuration

        This is the _second_ function being called after the Node has been instantiated.

        IMPORTANT: this function is run _before_ start of Twisted reactor!
        """
        config_source = None
        config_path = None

        # if the node hasn't been configured from XBR network, fallback to loading config from local config file
        if not self._config:
            default_filename = pkg_resources.resource_filename('crossbar', self.DEFAULT_CONFIG_PATH)
            with open(default_filename) as f:
                default_config = json.load(f)
            config_source, config_path = node.Node.load_config(self, configfile, default_config)
            self.log.info('Node configuration loaded from {config_source} ({config_path})',
                          config_source=hlid(config_source),
                          config_path=hlid(config_path))

        # Docker host integration
        if _HAS_DOCKER and self._config and 'controller' in self._config:
            self._enable_docker = self._config['controller'].get('enable_docker', False)

        return config_source, config_path

    def _watchdog(self):
        # on Linux, check that we start with sufficient system entropy
        entropy_avail = None
        if sys.platform.startswith('linux'):
            try:
                with open('/proc/sys/kernel/random/entropy_avail', 'r') as ent:
                    entropy_avail = int(ent.read())
                    # FIXME: my machine never has more than ~ 300 units available, 1000 seems a little optomistic!
                    if entropy_avail < 64:
                        self.log.warn('WARNING: insufficient entropy ({} bytes) available - try installing rng-tools'.format(entropy_avail))
            except PermissionError:
                # this happens when packaged as a snap: the code prevented from reading a location
                # # that is not allowed to a confined snap package
                entropy_avail = -1

        # check for at least 100MB free memory
        mem_avail = psutil.virtual_memory().available // 2 ** 20
        if mem_avail < 100:
            self.log.warn('WARNING: available memory dropped to {mem_avail} MB', mem_avail=mem_avail)

        self.log.trace('WATCHDOG: entropy_avail {entropy_avail} bytes, mem_avail {mem_avail} MB',
                       entropy_avail=entropy_avail, mem_avail=mem_avail)

    @inlineCallbacks
    def start(self, node_id=None):
        self.log.info('{note} [{method}]',
                      note=hl('Starting node (initialize edge-node personality) ..', color='green', bold=True),
                      method=hltype(FabricNode.start))

        # run watchdog at 5Hz
        self._watchdog_looper = LoopingCall(self._watchdog)
        self._watchdog_looper.start(.2)

        res = yield node.Node.start(self, node_id or self._node_id)
        return res

    @inlineCallbacks
    def boot(self, use_activation_file=True, use_default_fabric=False):
        self.log.info('Booting node {method}', method=hltype(FabricNode.boot))

        def reboot():
            self.stop(restart=True)

        # determine the transport configuration of the management uplink for this node
        fabric_transport_config = compute_mgmt_uplink_config(self.log, self._cbdir, self._config, reboot,
                                                             use_activation_file=use_activation_file,
                                                             use_default_fabric=use_default_fabric)

        # now finally, if we do have a transport configuration for the management uplink at this point,
        # then start the management uplink ..
        if fabric_transport_config:

            self._has_management_config = True

            url = fabric_transport_config['url']
            hostname = None
            if 'tls' in fabric_transport_config.get('endpoint', {}):
                hostname = fabric_transport_config['endpoint']['tls']['hostname']

            self._manager_runner = ApplicationRunner(
                url=url,
                realm=None,
                extra=None,
                ssl=optionsForClientTLS(hostname) if hostname else None,
            )

            def make(config):
                # extra info forwarded to CFC client session
                extra = {
                    'node': self,
                    'on_ready': Deferred(),
                    'on_exit': Deferred(),
                    'node_key': self._node_key,
                }

                @inlineCallbacks
                def on_ready_success(res):
                    try:
                        self._manager, self._management_realm, self._management_session_id, self._node_id, self._node_extra = res
                        if self._bridge_session:
                            try:
                                yield self._bridge_session.attach_manager(
                                    self._manager, self._management_realm, self._node_id)
                            except:
                                self.log.failure()
                            else:
                                while True:
                                    try:
                                        # we actually test the management uplink by calling a procedure on the master
                                        yield self._manager.call('crossbarfabriccenter.mrealm.get_status')
                                    except ApplicationError as e:
                                        if e.error == 'wamp.error.no_such_procedure':
                                            self.log.info('Could not get master status ("wamp.error.no_such_procedure") - retrying in 5 secs ..')
                                        else:
                                            self.log.failure()
                                        yield sleep(5)
                                    except:
                                        self.log.failure()
                                        self.log.info('Could not get master status - retrying in 5 secs ..')
                                        yield sleep(5)
                                    else:
                                        self.log.info(
                                            click.style(
                                                'Connected to Crossbar.io Master at management realm "{realm}", set node ID "{node_id}" (extra={node_extra}, session_id={session_id})',
                                                fg='green',
                                                bold=True),
                                            realm=self._management_realm,
                                            node_id=self._node_id,
                                            node_extra=self._node_extra,
                                            session_id=self._manager._session_id)

                                        # if the management uplink was successfully established and tested once, mark it so
                                        if not self._was_management_connected:
                                            self._was_management_connected = True
                                            self._management_url = url

                                        try:
                                            worker_ids = yield self._bridge_session.call(
                                                'crossbar.get_workers')
                                            for worker_id in worker_ids:
                                                yield self._bridge_session.call(
                                                    'crossbar.worker.{}.set_node_id'.format(worker_id),
                                                    self._node_id)
                                        except:
                                            self.log.warn(
                                                'INTERNAL ERROR: could not set node_id "{node_id}" after CFC connection was established',
                                                node_id=self._node_id)
                                            self.log.failure()

                                        break
                        else:
                            self.log.warn(
                                'Uplink Crossbar.io Master session established, but no bridge session setup!'
                            )
                    except Exception as e:
                        self.log.warn('error in on_ready: {}'.format(e))

                    # ok, we are connected to CFC and normally will be configurated programmatically from there.
                    # however, it is still possible to apply any local node configuration by setting
                    #
                    # node_extra:
                    # {
                    #    "on_start_apply_config", true
                    # }
                    #
                    # node_extra comes from CFC and has to be configured there (when the node is paired)
                    #
                    if self._node_extra:

                        # by default, apply local config (from a node configuration file, if there is one)
                        on_start_apply_config = self._node_extra.get('on_start_apply_config', True)

                        if on_start_apply_config:
                            if not self._local_config_applied:
                                self.log.info('Applying local node configuration (on_start_apply_config is enabled)')
                                yield self.boot_from_config(self._config)
                                self._local_config_applied = True
                            else:
                                self.log.info('Local node configuration was already applied - skipping')
                    else:
                        self.log.info('Skipping any local node configuration (no local config or on_start_apply_config is "off")')

                def on_ready_error(err):
                    if isinstance(
                            err.value,
                            ApplicationError) and err.value.error in ['fabric.auth-failed.node-unpaired', 'fabric.auth-failed.node-already-connected']:

                        if not self._displayed_pairing_message:
                            self._displayed_pairing_message = True
                            self.log.error(
                                click.style(err.value.error_message().upper(), fg='red', bold=True))

                        self.stop()
                    else:
                        self.log.error(click.style(
                            'Could not connect to CFC: {error}', fg='red', bold=True), error=err.value
                        )

                @inlineCallbacks
                def on_exit_success(reason):
                    if self._bridge_session:
                        try:
                            yield self._bridge_session.detach_manager()
                        except:
                            self.log.failure()
                        else:
                            self.log.debug(
                                'Disconnected from Crossbar.io Master for management realm "{realm}"',
                                realm=self._management_realm)
                    else:
                        self.log.warn(
                            'Uplink Crossbar.io Master session lost, but no bridge session setup!')

                    self._manager, self._management_realm, self._management_session_id, self._node_id, self._node_extra = None, None, None, None, None

                def on_exit_error(err):
                    print(err)

                extra['on_ready'].addCallbacks(on_ready_success, on_ready_error)
                extra['on_exit'].addCallbacks(on_exit_success, on_exit_error)

                config = ComponentConfig(extra=extra)
                session = NodeManagementSession(self._manager_runner, config)
                return session

            self.log.info('Connecting to Crossbar.io Master at {url} ..', url=url)

            yield self._manager_runner.run(make, start_reactor=False, auto_reconnect=True)

        else:
            # here, we don't have a management uplink :(
            self.log.info(
                hl('No management uplink configured (running unmanaged/single-node)',
                   color='red',
                   bold=True))
            self._has_management_config = False

            # nevertheless, now boot from local node config!
            yield self.boot_from_config(self._config)
            self._local_config_applied = True

    def _add_extra_controller_components(self, controller_config):
        extra = {
            'node_key': self._node_key,
            'controller_config': controller_config,
        }
        cfg = ComponentConfig(self._realm, extra=extra)
        self._bridge_session = NodeManagementBridgeSession(cfg)
        router = self._router_factory.get(self._realm)
        self._router_session_factory.add(self._bridge_session, router, authrole='trusted')

    def _set_shutdown_triggers(self, controller_options):
        if 'shutdown' in controller_options:
            self._node_shutdown_triggers = controller_options['shutdown']
            self.log.info("Using node shutdown triggers {triggers} from configuration", triggers=self._node_shutdown_triggers)
        else:
            # NODE_SHUTDOWN_ON_SHUTDOWN_REQUESTED
            # NODE_SHUTDOWN_ON_WORKER_EXIT
            # NODE_SHUTDOWN_ON_WORKER_EXIT_WITH_ERROR
            # NODE_SHUTDOWN_ON_LAST_WORKER_EXIT

            # in managed mode, a node - by default - only shuts down when explicitly asked to,
            # or upon a fatal error in the node controller
            self._node_shutdown_triggers = [checkconfig.NODE_SHUTDOWN_ON_SHUTDOWN_REQUESTED]

            self.log.info("Using default node shutdown triggers {triggers}", triggers=self._node_shutdown_triggers)

    def _add_worker_role(self, worker_auth_role, options):
        worker_role_config = {
            # each (native) worker is authenticated under a worker-specific authrole
            "name":
            worker_auth_role,
            "permissions": [
                # the worker requires these permissions to work:
                {
                    # management API provided by the worker. note that the worker management API is provided under
                    # the URI prefix "crossbar.worker.<worker_id>". note that the worker is also authenticated
                    # under authrole <worker_auth_role> on realm "crossbar"
                    "uri": worker_auth_role,
                    "match": "prefix",
                    "allow": {
                        "call": True,
                        "register": True,
                        "publish": True,
                        "subscribe": True
                    },
                    "disclose": {
                        "caller": True,
                        "publisher": True
                    },
                    "cache": True
                },
                {
                    # controller procedure called by the worker (to check for controller status)
                    "uri": "crossbar.get_status",
                    "match": "exact",
                    "allow": {
                        "call": True,
                        "register": False,
                        "publish": False,
                        "subscribe": False
                    },
                    "disclose": {
                        "caller": True,
                        "publisher": True
                    },
                    "cache": True
                }
            ]
        }
        # if configured to expose the controller connection within the worker (to make it available
        # in user code such as dynamic authenticators and router/container components), also add
        # permissions to actually use the (local) node management API
        if options.get('expose_controller', True):
            vendor_permissions = {
                "uri": "crossbar.",
                "match": "prefix",
                "allow": {
                    "call": True,
                    "register": False,
                    "publish": False,
                    "subscribe": True
                },
                "disclose": {
                    "caller": True,
                    "publisher": True
                },
                "cache": True
            }
            worker_role_config["permissions"].append(vendor_permissions)
            vendor_permissions = {
                "uri": "crossbarfabriccenter.",
                "match": "prefix",
                "allow": {
                    "call": True,
                    "register": True,
                    "publish": True,
                    "subscribe": True
                },
                "disclose": {
                    "caller": True,
                    "publisher": True
                },
                "cache": True
            }
            worker_role_config["permissions"].append(vendor_permissions)

        self._router_factory.add_role(self._realm, worker_role_config)

        self.log.info(
            'worker-specific role "{authrole}" added on node management router realm "{realm}" {func}',
            func=hltype(self._add_worker_role),
            authrole=hlid(worker_role_config['name']),
            realm=hlid(self._realm))

    def _extend_worker_args(self, args, options):
        if 'expose_shared' in options and options['expose_shared']:
            args.extend(['--expose_shared=true'])
        if 'expose_controller' in options and options['expose_controller']:
            args.extend(['--expose_controller=true'])

    @inlineCallbacks
    def _configure_native_worker_connections(self, worker_logname, worker_id, worker):
        # start connections (such as PostgreSQL database connection pools)
        # to run embedded in the router
        for connection in worker.get('connections', []):

            if 'id' in connection:
                connection_id = connection.pop('id')
            else:
                connection_id = 'connection{:03d}'.format(self._connection_no)
                self._connection_no += 1

            yield self._controller.call('crossbar.worker.{}.start_connection'.format(worker_id), connection_id, connection, options=CallOptions())
            self.log.info(
                "{logname}: connection '{connection}' started",
                logname=worker_logname,
                connection=connection_id,
            )

    @inlineCallbacks
    def _configure_native_worker_router(self, worker_logname, worker_id, worker):
        # setup db connection pool
        yield self._configure_native_worker_connections(worker_logname, worker_id, worker)

        # in this case, call the base class method _after_ above - because we want db connections
        # to be available when router/container components might start ..
        yield node.Node._configure_native_worker_router(self, worker_logname, worker_id, worker)

    @inlineCallbacks
    def _configure_native_worker_container(self, worker_logname, worker_id, worker):
        # setup db connection pool
        yield self._configure_native_worker_connections(worker_logname, worker_id, worker)

        # in this case, call the base class method _after_ above - because we want db connections
        # to be available when router/container components might start ..
        yield node.Node._configure_native_worker_container(self, worker_logname, worker_id, worker)

    @inlineCallbacks
    def _configure_native_worker_hostmonitor(self, worker_logname, worker_id, worker):
        # after the native worker has started, and the HostMonitor controller session
        # has attached to the local node router, we need to do the following, if we
        # want it to work _also_ from a local node config.json. driving from CFC
        # at run-time always works (also without the bits here)

        # 1. the native workers' common options are configured by calling into the worker
        yield self._configure_native_worker_common(worker_logname, worker_id, worker)

        # 2. the host monitor specific actions need to be done, namely, starting the monitoring
        monitoring_config = worker.get('monitor', None)
        yield self._controller.call('crossbar.worker.{}.start_monitoring'.format(worker_id), monitoring_config, options=CallOptions())

    @inlineCallbacks
    def _configure_native_worker_xbrmm(self, worker_logname, worker_id, worker):
        # 1. configure native workers' common options
        yield self._configure_native_worker_common(worker_logname, worker_id, worker)

        # 2. configure market makers (defined within the xbrmm worker)
        for maker in worker.get('makers', []):
            if 'id' in maker:
                maker_id = maker.pop('id')
            else:
                maker_id = 'maker{:03d}'.format(self._maker_no)
                self._maker_no += 1
            maker['id'] = maker_id
            yield self._controller.call('crossbar.worker.{}.start_market_maker'.format(worker_id), maker_id, maker, options=CallOptions())



#####################################################################################
#
#  Copyright (c) Crossbar.io Technologies GmbH
#  SPDX-License-Identifier: EUPL-1.2
#
#####################################################################################
"""
Server interop tests, making sure Crossbar's MQTT adapter responds the same as
other MQTT servers.
"""

import click
import attr
import sys

from collections import deque
from texttable import Texttable

from twisted.internet.protocol import Protocol, ClientFactory
from crossbar.bridge.mqtt.protocol import MQTTClientParser


@attr.s
class Frame(object):
    send = attr.ib()
    data = attr.ib()


class ConnectionLoss(object):
    send = False
    data = b""


@attr.s
class Result(object):
    name = attr.ib()
    success = attr.ib()
    reason = attr.ib()
    transcript = attr.ib()


@click.command()
@click.option("--host", help='Host address to test.')
@click.option("--port", type=int, help='Post of the host to test.')
def run(host, port):

    port = int(port)

    from . import interop_tests
    test_names = [x for x in dir(interop_tests) if x.startswith("test_")]

    tests = [getattr(interop_tests, test_name) for test_name in test_names]

    results = []
    with click.progressbar(tests, label="Running interop tests...") as _tests:
        for test in _tests:
            results.append(test(host, port))

    fmt_results = []
    for r in results:
        fmt_results.append((r.name, "True" if r.success else "False", r.reason if r.reason else "", r.transcript))

    t = Texttable()
    t.set_cols_width([20, 10, 80, 60])
    rows = [["Name", "Successful", "Reason", "Client Transcript"]]
    rows.extend(fmt_results)
    t.add_rows(rows)
    print(t.draw(), file=sys.__stdout__)

    failures = []
    for x in results:
        if not x.success:
            failures.append(False)

    if failures:
        sys.exit(len(failures))
    sys.exit(0)


class ReplayProtocol(Protocol):
    def __init__(self, factory):
        self.factory = factory
        self._record = deque(self.factory.record)
        self._waiting_for_nothing = None
        self._client = MQTTClientParser()

    def connectionMade(self):

        if self._record[0].send:
            to_send = self._record.popleft()
            if isinstance(to_send.data, bytes):
                self.transport.write(to_send.data)
            else:
                self.transport.write(to_send.data.serialise())

    def dataReceived(self, data):
        self.factory._timer.reset(7)

        got_data = self._client.data_received(data)
        self.factory.client_transcript.extend(got_data)

        if self._waiting_for_nothing:
            if data == b"":
                got_data.append(b"")
                self._waiting_for_nothing = None
            else:
                self.factory.reason = "Got unexpected data " + repr(got_data)
                self.factory.success = False
                self.factory.reactor.stop()
                return

        if len(self._record) > 0 and got_data:
            for x in got_data:
                reading = self._record.popleft()

                if x == reading.data:
                    pass
                elif isinstance(reading.data, list) and x in reading.data:
                    reading.data.remove(x)
                else:
                    self.factory.success = False
                    self.factory.reason = (x, reading.data)
                    self.factory.reactor.stop()
                    return

                if len(self._record) > 0:
                    while len(self._record) > 0 and self._record[0].send:
                        to_send = self._record.popleft()
                        if isinstance(to_send.data, bytes):
                            self.transport.write(to_send.data)
                        else:
                            self.transport.write(to_send.data.serialise())

                if isinstance(reading.data, list):
                    if reading.data:
                        self._record.appendleft(reading)

                if len(self._record) > 0:

                    # Then if we are supposed to wait...
                    if isinstance(self._record[0],
                                  Frame) and self._record[0].send is False and self._record[0].data == b"":

                        def wait():
                            self.dataReceived(b"")

                        self._waiting_for_nothing = self.factory.reactor.callLater(2, wait)
                        return

    def connectionLost(self, reason):
        if self.factory.reactor.running:
            if self._record and isinstance(self._record[0], ConnectionLoss):
                self.factory.success = True
            else:
                self.factory.success = False
                self.factory.reason = "Premature disconnection"
            self.factory.reactor.stop()


@attr.s
class ReplayClientFactory(ClientFactory):
    reactor = attr.ib()
    record = attr.ib()
    success = attr.ib(default=None)
    reason = attr.ib(default=None)
    protocol = ReplayProtocol
    noisy = False

    def buildProtocol(self, addr):

        self.client_transcript = []

        p = self.protocol(self)

        def disconnect():
            self.reason = "Timeout (remaining assertions were " + repr(p._record) + ")"
            self.reactor.stop()

        self._timer = self.reactor.callLater(7, disconnect)
        return p


if __name__ == "__main__":
    run()



##############################################################################
#
#                        Crossbar.io
#     Copyright (C) Crossbar.io Technologies GmbH. All rights reserved.
#
##############################################################################

import psutil

from crossbar.edge.worker.monitor._base import Monitor
from txaio import perf_counter_ns

__all__ = ('NetMonitor', )


class NetMonitor(Monitor):
    """
    Network monitoring via psutils.
    """

    ID = u'network'

    def poll(self):
        """
        Measure current stats value and return new stats.
        """
        hdata = Monitor.poll(self)

        start = perf_counter_ns()

        counters = {}
        io_counters = psutil.net_io_counters(True)
        for dev in io_counters:
            if dev.startswith('virbr') or dev.startswith('lo') or dev.startswith('docker'):
                continue
            counters[dev] = io_counters[dev]._asdict()

        hdata['net_io_counters'] = counters
        hdata['net_if_addrs'] = psutil.net_if_addrs()

        hdata[u'elapsed'] = perf_counter_ns() - start

        self._last_value = hdata

        return hdata



###############################################################################
#
# Crossbar.io Master
# Copyright (c) Crossbar.io Technologies GmbH. Licensed under EUPLv1.2.
#
###############################################################################

from typing import Dict
from collections.abc import Mapping
from pprint import pformat

import txaio

from crossbar.common import checkconfig
from crossbar.node.node import NodeOptions

from crossbar.personality import Personality as CrossbarPersonality
from crossbar.edge.personality import Personality as CrossbarFabricPersonality

from crossbar.master.node.node import FabricCenterNode

from crossbar.master.webservice import RouterWebServiceRegisterMe


def check_controller_fabric_center(personality, config):
    if not isinstance(config, Mapping):
        raise checkconfig.InvalidConfigException(
            "'fabric-center' in controller configuration must be a dictionary ({} encountered)\n\n".format(
                type(config)))

    for k in config:
        if k not in ['metering', 'auto_default_mrealm']:
            raise checkconfig.InvalidConfigException(
                "encountered unknown attribute '{}' in 'fabric-center' in controller configuration".format(k))

    if 'auto_default_mrealm' in config:
        auto_default_mrealm = config['auto_default_mrealm']
        checkconfig.check_dict_args(
            {
                'enabled': (False, [bool]),
                'watch_to_pair': (False, [str]),
                'watch_to_pair_pattern': (False, [str]),
                'write_pairing_file': (False, [bool]),
            }, auto_default_mrealm, "auto_default_mrealm configuration: {}".format(pformat(auto_default_mrealm)))

    if 'metering' in config:
        # "metering": {
        #     "period": 60,
        #     "submit": {
        #         "period": 120,
        #         "url": "http://localhost:7000",
        #         "timeout": 5,
        #         "maxerrors": 10
        #     }
        # }
        #
        # also possible to read the URL from an env var:
        #
        # "url": "${crossbar_METERING_URL}"
        #
        metering = config['metering']
        checkconfig.check_dict_args({
            'period': (False, [int, float]),
            'submit': (False, [Mapping]),
        }, metering, "metering configuration: {}".format(pformat(metering)))

        if 'submit' in metering:
            checkconfig.check_dict_args(
                {
                    'period': (False, [int, float]),
                    'url': (False, [str]),
                    'timeout': (False, [int, float]),
                    'maxerrors': (False, [int]),
                }, metering['submit'], "metering submit configuration: {}".format(pformat(metering['submit'])))

            if 'url' in metering['submit']:
                # allow to set value from environment variable
                metering['submit']['url'] = checkconfig.maybe_from_env('metering.submit.url',
                                                                       metering['submit']['url'])


def check_controller(personality, controller, ignore=[]):
    res = checkconfig.check_controller(personality, controller,
                                       ['fabric-center', 'enable_docker', 'blockchain', 'ipfs'] + ignore)

    if 'fabric-center' in controller:
        check_controller_fabric_center(personality, controller['fabric-center'])

    return res


def check_controller_options(personality, options, ignore=[]):
    return checkconfig.check_controller_options(personality, options, ignore)


class Personality(CrossbarFabricPersonality):

    log = txaio.make_logger()

    NAME = 'master'

    TEMPLATE_DIRS = [('crossbar', 'master/webservice/templates')] + CrossbarFabricPersonality.TEMPLATE_DIRS

    WEB_SERVICE_CHECKERS: Dict[str, object] = {
        'registerme': RouterWebServiceRegisterMe.check,
        **CrossbarPersonality.WEB_SERVICE_CHECKERS
    }

    WEB_SERVICE_FACTORIES: Dict[str, object] = {
        'registerme': RouterWebServiceRegisterMe,
        **CrossbarPersonality.WEB_SERVICE_FACTORIES
    }

    check_controller = check_controller
    check_controller_options = check_controller_options

    Node = FabricCenterNode
    NodeOptions = NodeOptions



#####################################################################################
#
#  Copyright (c) Crossbar.io Technologies GmbH
#  SPDX-License-Identifier: EUPL-1.2
#
#####################################################################################

from autobahn.twisted.resource import WebSocketResource

from crossbar.router.protocol import WampWebSocketServerFactory, WebSocketReverseProxyServerFactory
from crossbar.webservice.base import RouterWebService


class RouterWebServiceWebSocket(RouterWebService):
    """
    WAMP-WebSocket service.
    """
    @staticmethod
    def create(transport, path, config):
        websocket_factory = WampWebSocketServerFactory(transport._worker._router_session_factory, transport.cbdir,
                                                       config, transport.templates)

        # FIXME: Site.start/stopFactory should start/stop factories wrapped as Resources
        websocket_factory.startFactory()

        resource = WebSocketResource(websocket_factory)

        return RouterWebServiceWebSocket(transport, path, config, resource)


class RouterWebServiceWebSocketReverseProxy(RouterWebService):
    """
    Reverse WebSocket service.
    """
    @staticmethod
    def create(transport, path, config):
        ws_rproxy_factory = WebSocketReverseProxyServerFactory(transport._worker._reactor, config)
        ws_rproxy_factory.startFactory()
        resource = WebSocketResource(ws_rproxy_factory)
        return RouterWebServiceWebSocketReverseProxy(transport, path, config, resource)



###############################################################################
#
# Crossbar.io Shell
# Copyright (c) Crossbar.io Technologies GmbH. Licensed under EUPLv1.2.
#
###############################################################################

import os
from six.moves import configparser

# pair a node from a node public key from a local file:
#
# cbf pair node --realm "myrealm" --node "mynode" /var/local/crossbar/.crossbar/key.pub

# pair a node from a node public key served from a HTTP URL:
#
# cbf pair node --realm "myrealm" --node "mynode" http://localhost:9140/key.pub

from txaio import make_logger


class Profile(object):

    log = make_logger()

    def __init__(self,
                 name=None,
                 url=None,
                 reconnect=None,
                 debug=None,
                 realm=None,
                 role=None,
                 pubkey=None,
                 privkey=None,
                 tls_hostname=None,
                 tls_certificates=None):
        self.name = name
        self.url = url
        self.reconnect = reconnect
        self.debug = debug
        self.realm = realm
        self.role = role
        self.pubkey = pubkey
        self.privkey = privkey
        self.tls_hostname = tls_hostname
        self.tls_certificates = tls_certificates

    def __str__(self):
        return u'Profile(name={}, url={}, reconnect={}, debug={}, realm={}, role={}, pubkey={}, privkey={}, tls_hostname={}, tls_certificates={})'.format(
            self.name, self.url, self.reconnect, self.debug, self.realm, self.role, self.pubkey, self.privkey,
            self.tls_hostname, self.tls_certificates)

    @staticmethod
    def parse(name, items):
        url = None
        reconnect = None
        debug = None
        realm = None
        role = None
        pubkey = None
        privkey = None
        tls_hostname = None
        tls_certificates = None
        for k, v in items:
            if k == 'url':
                url = str(v)
            elif k == 'reconnect':
                reconnect = int(v)
            elif k == 'debug':
                debug = bool(v)
            elif k == 'realm':
                realm = str(v)
            elif k == 'role':
                role = str(v)
            elif k == 'pubkey':
                pubkey = str(v)
            elif k == 'privkey':
                privkey = str(v)
            elif k == 'tls_hostname':
                tls_hostname = str(v)
            elif k == 'tls_certificates':
                tls_certificates = [x.strip() for x in str(v).split(',')]
            else:
                # skip unknown attribute
                Profile.log.warn('unprocessed config attribute "{}"'.format(k))

        return Profile(name, url, reconnect, debug, realm, role, pubkey, privkey, tls_hostname, tls_certificates)


class UserConfig(object):

    log = make_logger()

    def __init__(self, config_path):
        self._config_path = os.path.abspath(config_path)

        config = configparser.ConfigParser()
        config.read(config_path)

        self.config = config

        profiles = {}
        for profile_name in config.sections():
            profile = Profile.parse(profile_name, config.items(profile_name))
            profiles[profile_name] = profile
            self.log.info('Profile "{profile_name}" parsed: {profile}', profile_name=profile_name, profile=profile)

        self.profiles = profiles

        self.log.info('Profiles loaded for: {profiles}', profiles=sorted(self.profiles.keys()))



###############################################################################
#
# Crossbar.io Master
# Copyright (c) Crossbar.io Technologies GmbH. Licensed under EUPLv1.2.
#
###############################################################################

import uuid
from pprint import pformat
from typing import Optional, List, Dict, Tuple

import numpy as np

from autobahn import wamp
from autobahn.wamp.exception import ApplicationError
from autobahn.wamp.types import CallDetails, PublishOptions, RegisterOptions

from crossbar.common import checkconfig
from crossbar.webservice import archive, wap
from crossbar._util import hl, hlid, hltype, hlval, get_free_tcp_port
from cfxdb.mrealm import WebCluster, WebClusterNodeMembership, WebService
from cfxdb.mrealm import cluster
from cfxdb.mrealm.application_realm import ApplicationRealmStatus

import txaio

txaio.use_twisted()
from txaio import time_ns, sleep, make_logger  # noqa
from twisted.internet.defer import inlineCallbacks
from twisted.internet.task import LoopingCall


class WebClusterMonitor(object):
    """
    Background monitor running periodically in the master node to monitor, check and apply
    necessary actions for web clusters.

    The monitor is started when a web cluster is started.
    """
    log = make_logger()

    def __init__(self, manager, webcluster_oid, interval=10.):
        self._manager = manager
        self._webcluster_oid = webcluster_oid
        self._interval = interval

        self._loop = None
        self._check_and_apply_in_progress = False

        self._workers = {}

    @property
    def is_started(self):
        """

        :return:
        """
        return self._loop is not None and self._loop.running

    def start(self):
        """

        :return:
        """
        assert self._loop is None

        self._loop = LoopingCall(self._check_and_apply)
        self._loop.start(self._interval)

    def stop(self):
        """

        :return:
        """
        assert self._loop is not None

        self._loop.stop()
        self._loop = None
        self._check_and_apply_in_progress = False

    def get_cluster_workers(self, filter_online: bool = False) -> List[Tuple[str, str]]:
        """

        :param filter_online:
        :return:
        """
        res = []
        for node_oid, worker_id in self._workers:
            if not filter_online or self._workers[(node_oid, worker_id)]['status'] == 'started':
                res.append((node_oid, worker_id))
        return res

    @inlineCallbacks
    def _check_and_apply(self):
        if self._check_and_apply_in_progress:
            # we prohibit running the iteration multiple times concurrently. this might
            # happen when the iteration takes longer than the interval the monitor is set to
            self.log.info('{func} {action} for webcluster {webcluster} skipped! check & apply already in progress.',
                          action=hl('check & apply run skipped', color='red', bold=True),
                          func=hltype(self._check_and_apply),
                          webcluster=hlid(self._webcluster_oid))
            return
        else:
            self.log.info('{func} {action} for webcluster {webcluster} ..',
                          action=hl('check & apply run started', color='green', bold=True),
                          func=hltype(self._check_and_apply),
                          webcluster=hlid(self._webcluster_oid))
            self._check_and_apply_in_progress = True

        is_running_completely = True
        workers = {}
        try:
            # get all active (non-standby) nodes added to the webcluster
            active_memberships = []
            with self._manager.db.begin() as txn:
                # the webcluster itself
                webcluster = self._manager.schema.webclusters[txn, self._webcluster_oid]

                if webcluster.status in [cluster.STATUS_STARTING, cluster.STATUS_RUNNING]:
                    # the node memberships in the webcluster
                    active_memberships = [
                        m for m in self._manager.schema.webcluster_node_memberships.select(
                            txn, from_key=(webcluster.oid, uuid.UUID(bytes=b'\0' * 16)), return_keys=False)
                    ]

            for membership in active_memberships:
                node_oid = str(membership.node_oid)

                if membership.standby:
                    self.log.debug('{func} Web cluster node {node_oid} is configured for standby',
                                   func=hltype(self._check_and_apply),
                                   node_oid=hlid(node_oid))
                    continue

                # node run-time information, as maintained here in our master view of the external world
                node = self._manager._session.nodes.get(node_oid, None)

                if node and node.status == 'online':
                    self.log.debug('{func} Ok, web cluster node {node_oid} is running!',
                                   func=hltype(self._check_and_apply),
                                   node_oid=hlid(node_oid))

                    # we expect "parallel" workers to run on this node ..
                    for worker_index in range(membership.parallel or 1):

                        # run-time ID of web cluster worker, eg "clwrk-a276279d-5"
                        worker_id = 'cpw-{}-{}'.format(str(webcluster.oid)[:8], worker_index)

                        self.log.debug(
                            '{func} Performing checks for configured proxy worker {worker_index}/{parallel} [{worker_id}] ..',
                            func=hltype(self._check_and_apply),
                            worker_index=hlid(worker_index + 1),
                            worker_id=hlid(worker_id),
                            parallel=hlid(membership.parallel))

                        # worker run-time information (obtained by calling into the live node)
                        worker = None
                        try:
                            worker = yield self._manager._session.call('crossbarfabriccenter.remote.node.get_worker',
                                                                       node_oid, worker_id)
                        except ApplicationError as e:
                            if e.error != 'crossbar.error.no_such_worker':
                                # anything but "no_such_worker" is unexpected (and fatal)
                                raise
                            self.log.info(
                                'No Web cluster worker {worker_id} currently running on node {node_oid}: starting worker ..',
                                node_oid=hlid(node_oid),
                                worker_id=hlid(worker_id))
                        else:
                            self.log.debug(
                                '{func} Ok, web cluster worker {worker_id} already running on node {node_oid}!',
                                func=hltype(self._check_and_apply),
                                node_oid=hlid(node_oid),
                                worker_id=hlid(worker_id))

                        # if there isn't a worker running (with worker ID as we expect) already,
                        # start a new proxy worker ..
                        if not worker:
                            worker_options = None
                            try:
                                worker_started = yield self._manager._session.call(
                                    'crossbarfabriccenter.remote.node.start_worker', node_oid, worker_id, 'proxy',
                                    worker_options)
                                worker = yield self._manager._session.call(
                                    'crossbarfabriccenter.remote.node.get_worker', node_oid, worker_id)
                                self.log.info(
                                    '{func} Web cluster worker {worker_id} started on node {node_oid} [{worker_started}]',
                                    func=hltype(self._check_and_apply),
                                    node_oid=hlid(node_oid),
                                    worker_id=hlid(worker_id),
                                    worker_started=worker_started)
                            except:
                                self.log.failure()
                                is_running_completely = False

                        # we can only continue with transport(s) when we now have a worker started already
                        if worker:
                            transport = None

                            # FIXME: currently, we only have 1 transport on a web cluster worker (which is named "primary")
                            transport_id = 'primary'
                            try:
                                transport = yield self._manager._session.call(
                                    'crossbarfabriccenter.remote.proxy.get_proxy_transport', node_oid, worker_id,
                                    transport_id)
                            except ApplicationError as e:
                                if e.error != 'crossbar.error.no_such_object':
                                    # anything but "no_such_object" is unexpected (and fatal)
                                    raise
                                self.log.info(
                                    '{func} No Transport {transport_id} currently running for Web cluster worker {worker_id}: starting transport ..',
                                    func=hltype(self._check_and_apply),
                                    worker_id=hlid(worker_id),
                                    transport_id=hlid(transport_id))
                            else:
                                self.log.debug(
                                    '{func} Ok, transport {transport_id} already running on Web cluster worker {worker_id}',
                                    func=hltype(self._check_and_apply),
                                    worker_id=hlid(worker_id),
                                    transport_id=hlid(transport_id))

                            # if there isn't a transport started (with transport ID as we expect) already,
                            # start a new transport ..
                            if not transport:
                                transport_config = {
                                    'id': transport_id,
                                    'type': 'web',
                                    'endpoint': {
                                        'type': 'tcp',
                                        'port':
                                        int(webcluster.tcp_port) if webcluster.tcp_port else get_free_tcp_port(),
                                        'shared': webcluster.tcp_shared is True,
                                    },
                                    'paths': {},
                                    'options': {
                                        'access_log': webcluster.http_access_log is True,
                                        'display_tracebacks': webcluster.http_display_tracebacks is True,
                                        'hsts': webcluster.http_hsts is True,
                                    }
                                }
                                if webcluster.tcp_interface:
                                    transport_config['endpoint']['interface'] = webcluster.tcp_interface
                                if webcluster.tcp_backlog:
                                    transport_config['endpoint']['backlog'] = webcluster.tcp_backlog
                                if webcluster.http_hsts_max_age:
                                    transport_config['options']['hsts_max_age'] = webcluster.http_hsts_max_age
                                if webcluster.http_client_timeout:
                                    transport_config['options']['client_timeout'] = webcluster.http_client_timeout

                                try:
                                    transport_started = yield self._manager._session.call(
                                        'crossbarfabriccenter.remote.proxy.start_proxy_transport', node_oid, worker_id,
                                        transport_id, transport_config)
                                    transport = yield self._manager._session.call(
                                        'crossbarfabriccenter.remote.proxy.get_proxy_transport', node_oid, worker_id,
                                        transport_id)
                                    self.log.info(
                                        '{func} Transport {transport_id} started on Web cluster worker {worker_id} [{transport_started}]',
                                        func=hltype(self._check_and_apply),
                                        worker_id=hlid(worker_id),
                                        transport_id=hlid(transport_id),
                                        transport_started=transport_started)
                                except:
                                    self.log.failure()
                                    is_running_completely = False

                            # we can only continue with web services when we now have a transport started already
                            if transport:
                                # collect all web services defined for our (one) web transport, and collect
                                # in a path->webservice map
                                webservices = {}
                                with self._manager.db.begin() as txn:
                                    for webservice_oid in self._manager.schema.idx_webcluster_webservices.select(
                                            txn, from_key=(webcluster.oid, uuid.UUID(bytes=b'\0' * 16)),
                                            return_keys=False):
                                        webservice = self._manager.schema.webservices[txn, webservice_oid]
                                        if webservice:
                                            webservices[webservice.path] = webservice
                                        else:
                                            self.log.warn('No webservice object found for oid {webservice_oid}',
                                                          webservice_oid=webservice_oid)

                                for path, webservice in webservices.items():
                                    service = None
                                    try:
                                        service = yield self._manager._session.call(
                                            'crossbarfabriccenter.remote.proxy.get_web_transport_service', node_oid,
                                            worker_id, transport_id, path)
                                    except ApplicationError as e:
                                        # anything but "not_running" is unexpected (and fatal)
                                        if e.error != 'crossbar.error.not_running':
                                            raise
                                        self.log.info(
                                            '{func} No Web service currently running on path "{path}" for Web cluster worker {worker_id} web transport {transport_id}: starting web service ..',
                                            func=hltype(self._check_and_apply),
                                            path=hlval(path),
                                            worker_id=hlid(worker_id),
                                            transport_id=hlid(transport_id))
                                    else:
                                        self.log.debug(
                                            '{func} Ok, web service on path "{path}" is already running for Web cluster worker {worker_id} web transport {transport_id}',
                                            func=hltype(self._check_and_apply),
                                            path=hlval(path),
                                            worker_id=hlid(worker_id),
                                            transport_id=hlid(transport_id))

                                    if not service:
                                        webservice_config = webservice.marshal()
                                        webservice_config.pop('oid', None)
                                        webservice_config.pop('label', None)
                                        webservice_config.pop('description', None)
                                        webservice_config.pop('tags', None)
                                        webservice_config.pop('cluster_oid', None)
                                        webservice_config.pop('path', None)

                                        # FIXME: this shouldn't be there (but should be cluster_oid)
                                        webservice_config.pop('webcluster_oid', None)
                                        try:
                                            webservice_started = yield self._manager._session.call(
                                                'crossbarfabriccenter.remote.proxy.start_web_transport_service',
                                                node_oid, worker_id, transport_id, path, webservice_config)
                                            self.log.info(
                                                '{func} Web service started on transport {transport_id} and path "{path}" [{webservice_started}]',
                                                func=hltype(self._check_and_apply),
                                                transport_id=hlid(transport_id),
                                                path=hlval(path),
                                                webservice_started=webservice_started)
                                        except:
                                            self.log.failure()
                                            is_running_completely = False

                            with self._manager.db.begin() as txn:
                                for arealm_oid in self._manager.schema.idx_arealm_by_webcluster.select(
                                        txn,
                                        from_key=(webcluster.oid, ''),
                                        to_key=(uuid.UUID(int=int(webcluster.oid) + 1), ''),
                                        return_keys=False):
                                    arealm = self._manager.schema.arealms[txn, arealm_oid]
                                    if arealm and arealm.status == ApplicationRealmStatus.RUNNING and arealm.workergroup_oid:
                                        self.log.debug(
                                            '{func} node {node_id} - worker {worker_id} - webcluster "{webcluster_name}": backend router workergroup {workergroup_oid} is associated with this frontend web cluster for application realm "{arealm_name}"',
                                            func=hltype(self._check_and_apply),
                                            node_id=hlval(str(node.node_id)),
                                            worker_id=hlval(worker_id),
                                            webcluster_name=hlval(webcluster.name),
                                            arealm_name=hlval(arealm.name),
                                            arealm_oid=hlid(arealm_oid),
                                            workergroup_oid=hlid(arealm.workergroup_oid))

                        wk = (node_oid, worker['id'])
                        workers[wk] = worker
                else:
                    self.log.warn('{func} Web cluster node {node_oid} not running [status={status}]',
                                  func=hltype(self._check_and_apply),
                                  node_oid=hlid(node_oid),
                                  status=hl(node.status if node else 'offline'))
                    is_running_completely = False

            if webcluster.status in [cluster.STATUS_STARTING] and is_running_completely:
                with self._manager.db.begin(write=True) as txn:
                    webcluster = self._manager.schema.webclusters[txn, self._webcluster_oid]
                    webcluster.status = cluster.STATUS_RUNNING
                    webcluster.changed = time_ns()
                    self._manager.schema.webclusters[txn, webcluster.oid] = webcluster

                webcluster_started = {
                    'oid': str(webcluster.oid),
                    'status': cluster.STATUS_BY_CODE[webcluster.status],
                    'changed': webcluster.changed,
                }
                yield self._manager._session.publish('{}.on_webcluster_started'.format(self._manager._prefix),
                                                     webcluster_started,
                                                     options=self._manager._PUBOPTS)
        except:
            self.log.failure()

        self._workers = workers
        for node_oid, worker_id in self._workers:
            worker = self._workers[(node_oid, worker_id)]
            if worker:
                status = worker['status'].upper()
            else:
                status = 'MISSING'
            self.log.info(
                '{func} webcluster {webcluster_oid} worker {worker_id} on node {node_oid} has status {status}',
                func=hltype(self._check_and_apply),
                worker_id=hlid(worker_id),
                node_oid=hlid(node_oid),
                webcluster_oid=hlid(self._webcluster_oid),
                status=hlval(status))

        if is_running_completely:
            color = 'green'
            action = 'check & apply run completed successfully'
        else:
            color = 'red'
            action = 'check & apply run finished with problems left'

        self._check_and_apply_in_progress = False
        self.log.info('{func} {action} for webcluster {webcluster}!',
                      action=hl(action, color=color, bold=True),
                      func=hltype(self._check_and_apply),
                      webcluster=hlid(self._webcluster_oid))


class WebClusterManager(object):
    """
    Manages Web clusters, which runs Crossbar.io Web transport listening
    endpoints on many (frontend) workers over many nodes using applying
    a shared, common transport definition, such as regarding the Web services
    configured on URL paths of the Web transport.
    """
    log = make_logger()

    # publication options for management API events
    _PUBOPTS = PublishOptions(acknowledge=True)

    # map of allowed web services, see also crossbar.personality.Personality.WEB_SERVICE_CHECKERS
    _WEB_SERVICE_CHECKERS = {
        # none
        'path': checkconfig.check_web_path_service_path,
        'redirect': checkconfig.check_web_path_service_redirect,
        # resource
        'reverseproxy': checkconfig.check_web_path_service_reverseproxy,
        'nodeinfo': checkconfig.check_web_path_service_nodeinfo,
        'json': checkconfig.check_web_path_service_json,
        'cgi': checkconfig.check_web_path_service_cgi,
        'wsgi': checkconfig.check_web_path_service_wsgi,
        'static': checkconfig.check_web_path_service_static,
        'websocket': checkconfig.check_web_path_service_websocket,
        'websocket-reverseproxy': checkconfig.check_web_path_service_websocket_reverseproxy,
        # longpoll
        'caller': checkconfig.check_web_path_service_caller,
        'publisher': checkconfig.check_web_path_service_publisher,
        # webhook
        'archive': archive.RouterWebServiceArchive.check,
        'wap': wap.RouterWebServiceWap.check,
    }

    def __init__(self, session, globaldb, globalschema, db, schema, reactor=None):
        """

        :param session: Backend of user created management realms.
        :type session: :class:`crossbar.master.mrealm.controller.MrealmController`

        :param globaldb: Global database handle.
        :type globaldb: :class:`zlmdb.Database`

        :param globalschema: Global database schema.
        :type globalschema: :class:`cfxdb.globalschema.GlobalSchema`

        :param db: Management realm database handle.
        :type db: :class:`zlmdb.Database`

        :param schema: Management realm database schema.
        :type schema: :class:`cfxdb.mrealmschema.MrealmSchema`
        """
        self._session = session

        # ContainerController
        self._worker = session.config.controller

        # node personality
        self._personality = self._worker.personality

        # twisted reactor
        self._reactor = self._worker._reactor
        # if not reactor:
        #     from twisted.internet import reactor
        # self._reactor = reactor

        # global database handle & schema

        # global (node level) database handle & schema
        self.gdb = globaldb
        self.gschema = globalschema

        # mrealm database handle & schema
        self.db = db
        self.schema = schema

        # will be set in session.register
        self._prefix = None

        # the management realm OID this webcluster manager operates for
        self._mrealm_oid = session._mrealm_oid

        # filled when started
        self._started = None

        # webcluster monitors, containing a map, for every webcluster in state STARTING or RUNNING
        # with objects of class WebClusterMonitor
        self._monitors = {}

    def get_webcluster_workers(self, webcluster_oid, filter_online=False):
        if webcluster_oid in self._monitors:
            return self._monitors[webcluster_oid].get_cluster_workers(filter_online=filter_online)

    @inlineCallbacks
    def start(self, prefix):
        """
        Start the Web-cluster manager.

        :return:
        """
        assert self._started is None, 'cannot start web cluster manager - already running!'
        assert self._prefix is None

        self._started = time_ns()

        # crossbarfabriccenter.mrealm.webcluster
        self._prefix = prefix[:-1] if prefix.endswith('.') else prefix

        regs = yield self._session.register(self,
                                            prefix='{}.'.format(self._prefix),
                                            options=RegisterOptions(details_arg='details'))
        procs = [reg.procedure for reg in regs]
        self.log.debug(
            'Web cluster manager {api} registered management procedures using prefix "{prefix}" [{func}]:\n\n{procs}\n',
            api=hl('Web cluster manager API', color='green', bold=True),
            func=hltype(self.start),
            prefix=hlval(self._prefix),
            procs=hl(pformat(procs), color='white', bold=True))

        # start all web cluster monitors ..
        cnt_started = 0
        cnt_skipped = 0
        dl = []
        with self.db.begin() as txn:
            webcluster_oids = self.schema.webclusters.select(txn, return_values=False)
            for webcluster_oid in webcluster_oids:
                webcluster = self.schema.webclusters[txn, webcluster_oid]
                if webcluster.status in [cluster.STATUS_STARTING, cluster.STATUS_RUNNING]:
                    monitor = WebClusterMonitor(self, webcluster_oid)
                    dl.append(monitor.start())
                    assert webcluster_oid not in self._monitors
                    self._monitors[webcluster_oid] = monitor
                    cnt_started += 1
                    self.log.info('{func}(prefix="{prefix}"): {action} for web cluster {webcluster_oid} in {status})',
                                  action=hl('cluster monitor started', color='green', bold=True),
                                  prefix=hlval(prefix),
                                  func=hltype(self.start),
                                  webcluster_oid=hlid(webcluster_oid),
                                  status=hlval(webcluster.status))
                else:
                    cnt_skipped += 1
                    self.log.info(
                        '{func}(prefix="{prefix}"): {action} for web cluster {webcluster_oid} in status {status}',
                        action=hl('cluster monitor skipped', color='green', bold=True),
                        prefix=hlval(prefix),
                        func=hltype(self.start),
                        webcluster_oid=hlid(webcluster_oid),
                        status=hlval(webcluster.status))
        self.log.info(
            'Web cluster manager has started monitors for {cnt_started} clusters ({cnt_skipped} skipped) [{func}]',
            cnt_started=hlval(cnt_started),
            cnt_skipped=hlval(cnt_skipped),
            func=hltype(self.start))

        self.log.info('Web cluster manager ready for management realm {mrealm_oid}! [{func}]',
                      mrealm_oid=hlid(self._mrealm_oid),
                      func=hltype(self.start))

        # return txaio.gather(dl)

    @inlineCallbacks
    def stop(self):
        """
        Stop the (currently running) Web-cluster manager.

        :return:
        """
        assert self._started > 0, 'cannot stop web cluster manager - currently not running!'

        # stop all web cluster monitors ..
        dl = []
        for webcluster_oid, webcluster_monitor in self._monitors.items():
            dl.append(webcluster_monitor.stop())
            del self._monitors[webcluster_oid]
        self._started = None
        self.log.info(
            'Ok, web cluster manager for management realm {mrealm_oid} stopped ({cnt_stopped} monitors stopped) [{func}]',
            mrealm_oid=hlid(self._mrealm_oid),
            cnt_stopped=len(dl),
            func=hltype(self.start))

        # return txaio.gather(dl)

    @wamp.register(None, check_types=True)
    def list_webclusters(self,
                         return_names: Optional[bool] = False,
                         details: Optional[CallDetails] = None) -> List[str]:
        """
        Returns list of web clusters defined. Detail information for a web cluster
        can be retrieved using :meth:`crossbar.master.cluster.webcluster.WebClusterManager.get_webcluster`.

        :param return_names: Return webcluster names instead of  object IDs

        :returns: List of WebCluster UUIDs (or names). For example:

            .. code-block:: json

                [
                    "4917ca20-acc5-497a-9801-b53db5db4d89"
                ]

            or with ``return_names``  set:

            .. code-block:: json

                [
                    "cluster1"
                ]
        """
        assert details is None or isinstance(details, CallDetails)

        self.log.info('{func}(details={details})', func=hltype(self.list_webclusters), details=details)

        with self.db.begin() as txn:
            if return_names:
                webclusters = self.schema.webclusters.select(txn, return_keys=False)
                if webclusters:
                    return sorted([webcluster.name for webcluster in webclusters])
                else:
                    return []
            else:
                webcluster_oids = self.schema.webclusters.select(txn, return_values=False)
                if webcluster_oids:
                    # we now have a list of uuid.UUID objects: convert to strings
                    return [str(oid) for oid in webcluster_oids]
                else:
                    return []

    @wamp.register(None, check_types=True)
    def get_webcluster(self, webcluster_oid: str, details: Optional[CallDetails] = None) -> dict:
        """
        Return configuration and run-time status information for a web cluster (by object ID).

        :param routercluster_oid: Object ID of the web cluster to return.

        :return: Web cluster definition. For example:

            .. code-block:: json

                {
                    "changed": 1598388779771813358,
                    "description": null,
                    "http_access_log": null,
                    "http_client_timeout": null,
                    "http_display_tracebacks": null,
                    "http_hsts": null,
                    "http_hsts_max_age": null,
                    "label": null,
                    "name": "cluster1",
                    "oid": "4917ca20-acc5-497a-9801-b53db5db4d89",
                    "owner_oid": "8d6e3068-900a-4fa8-a6f5-0828c8d0ee24",
                    "status": "STOPPED",
                    "tags": null,
                    "tcp_backlog": null,
                    "tcp_interface": null,
                    "tcp_port": 8080,
                    "tcp_shared": true,
                    "tcp_version": null,
                    "tls_ca_certificates": null,
                    "tls_certificate": null,
                    "tls_chain_certificates": null,
                    "tls_ciphers": null,
                    "tls_dhparam": null,
                    "tls_key": null
                }
        """
        assert type(webcluster_oid) == str
        assert details is None or isinstance(details, CallDetails)

        self.log.info('{func}(webcluster_oid={webcluster_oid}, details={details})',
                      func=hltype(self.get_webcluster),
                      webcluster_oid=hlid(webcluster_oid),
                      details=details)

        try:
            webcluster_oid_ = uuid.UUID(webcluster_oid)
        except Exception as e:
            raise ApplicationError('wamp.error.invalid_argument', 'invalid webcluster_oid: {}'.format(str(e)))

        with self.db.begin() as txn:
            webcluster = self.schema.webclusters[txn, webcluster_oid_]

        if webcluster:
            return webcluster.marshal()
        else:
            raise ApplicationError('crossbar.error.no_such_object',
                                   'no webcluster with oid {}'.format(webcluster_oid_))

    @wamp.register(None, check_types=True)
    def get_webcluster_by_name(self, webcluster_name: str, details: Optional[CallDetails] = None) -> dict:
        """
        Return configuration and run-time status information for a web cluster (by name).

        See also the corresponding procedure :meth:`crossbar.master.cluster.webcluster.WebClusterManager.get_webcluster`
        which returns the same information, given and object ID rather than name.

        :param webcluster_name: Name of the web cluster to return.

        :return: Web cluster definition.
        """
        assert type(webcluster_name) == str
        assert details is None or isinstance(details, CallDetails)

        self.log.info('{func}(webcluster_name="{webcluster_name}", details={details})',
                      func=hltype(self.get_webcluster_by_name),
                      webcluster_name=hlid(webcluster_name),
                      details=details)

        with self.db.begin() as txn:
            webcluster_oid = self.schema.idx_webclusters_by_name[txn, webcluster_name]
            if not webcluster_oid:
                raise ApplicationError('crossbar.error.no_such_object',
                                       'no webcluster named {}'.format(webcluster_name))

            webcluster = self.schema.webclusters[txn, webcluster_oid]
            assert webcluster

        return webcluster.marshal()

    @wamp.register(None, check_types=True)
    async def create_webcluster(self, webcluster: dict, details: Optional[CallDetails] = None) -> dict:
        """
        Create a new web cluster definition ``webcluster``:

        Web listening endpoint:

        - ``tcp_version``: IP version, either 4 for 6
        - ``tcp_port``: IP listening port
        - ``tcp_shared``: enable TCP port sharing
        - ``tcp_interface``: listen on this interface
        - ``tcp_backlog``: TCP accept backlog queue size

        Web endpoint TLS configuration:

        - ``tls_key``: TLS server private key to use
        - ``tls_certificate``: TLS server certificate to use
        - ``tls_chain_certificates``: TLS certificate chain
        - ``tls_dhparam``: DH parameter file
        - ``tls_ciphers``: Ciphers list
        - ``tls_ca_certificates``: CA certificates to use

        Web transport options:

        - ``http_client_timeout``: HTTP client inactivity timeout
        - ``http_hsts``: enable HTTP strict transport security (HSTS)
        - ``http_hsts_max_age``: HSTS maximum age to announce
        - ``http_access_log``: enable Web request access logging
        - ``http_display_tracebacks``: enable tracebacks when running into Web errors

        :procedure: ``crossbarfabriccenter.mrealm.cluster.create_webcluster`` URI of WAMP procedure to call.
        :event: ``crossbarfabriccenter.mrealm.cluster.on_webcluster_created`` WAMP event published once the
            web cluster has been created.
        :error: ``wamp.error.invalid_configuration`` WAMP error returned when the web cluster
            configuration provided has a problem.
        :error: ``wamp.error.not_authorized`` WAMP error returned when the user is currently not allowed
            to created (another) web cluster.
        :error: ``crossbar.error.already_exists`` WAMP error returned when a web cluster named as contained
            in the configuration already exists.

        :param webcluster: Web cluster settings. For example:

            .. code-block:: json

                {
                    "name": "cluster1"
                }


        :return: Web cluster creation information. For example:

            .. code-block:: json

                {
                    "changed": 1598388333642427113,
                    "description": null,
                    "http_access_log": null,
                    "http_client_timeout": null,
                    "http_display_tracebacks": null,
                    "http_hsts": null,
                    "http_hsts_max_age": null,
                    "label": null,
                    "name": "cluster1",
                    "oid": "96e3d9a6-3e88-4205-8eec-2e7c338b2620",
                    "owner_oid": "c10a7e49-cea6-47ce-a003-74f7196d1763",
                    "status": "STOPPED",
                    "tags": null,
                    "tcp_backlog": null,
                    "tcp_interface": null,
                    "tcp_port": 8080,
                    "tcp_shared": true,
                    "tcp_version": null,
                    "tls_ca_certificates": null,
                    "tls_certificate": null,
                    "tls_chain_certificates": null,
                    "tls_ciphers": null,
                    "tls_dhparam": null,
                    "tls_key": null
                }
        """
        assert type(webcluster) == dict
        assert details is None or isinstance(details, CallDetails)

        self.log.info('{func}(webcluster="{webcluster}", details={details})',
                      func=hltype(self.create_webcluster),
                      webcluster=webcluster,
                      details=details)

        try:
            obj = WebCluster.parse(webcluster)
        except Exception as e:
            raise ApplicationError('wamp.error.invalid_configuration',
                                   'could not parse web cluster configuration ({})'.format(e))

        if obj.name is None:
            raise ApplicationError('wamp.error.invalid_configuration', 'missing "name" in web cluster configuration')
        else:
            if not checkconfig._CONFIG_ITEM_ID_PAT.match(obj.name):
                raise ApplicationError(
                    'wamp.error.invalid_configuration',
                    'invalid name "{}" in web cluster configuration (must match {})'.format(
                        obj.name, checkconfig._CONFIG_ITEM_ID_PAT_STR))

        obj.oid = uuid.uuid4()
        obj.status = cluster.STATUS_BY_NAME['STOPPED']
        obj.changed = np.datetime64(time_ns(), 'ns')

        if details and details.caller_authid:
            with self.gdb.begin() as txn:
                caller_oid = self.gschema.idx_users_by_email[txn, details.caller_authid]
                if not caller_oid:
                    raise ApplicationError('wamp.error.no_such_principal',
                                           'no user found for authid "{}"'.format(details.caller_authid))
            obj.owner_oid = caller_oid
        else:
            raise ApplicationError('wamp.error.no_such_principal', 'cannot map user - no caller authid available')

        with self.db.begin(write=True) as txn:
            if self.schema.idx_webclusters_by_name[txn, obj.name]:
                raise ApplicationError('crossbar.error.already_exists',
                                       'duplicate name "{}" in web cluster configuration'.format(obj.name))

            self.schema.webclusters[txn, obj.oid] = obj

        self.log.info('new WebCluster object stored in database:\n{obj}', obj=obj)

        res_obj = obj.marshal()

        await self._session.publish('{}.on_webcluster_created'.format(self._prefix), res_obj, options=self._PUBOPTS)

        self.log.info('Management API event <on_webcluster_created> published:\n{res_obj}', res_obj=res_obj)

        return res_obj

    @wamp.register(None, check_types=True)
    async def delete_webcluster(self, webcluster_oid: str, details: Optional[CallDetails] = None) -> dict:
        """
        Delete an web router cluster definition. The web cluster must be in status ``"STOPPED"``.


        :procedure: ``crossbarfabriccenter.webcluster.delete_webcluster`` URI of WAMP procedure to call.
        :event: ``crossbarfabriccenter.webcluster.on_webcluster_deleted`` WAMP event published once the web cluster has been deleted.
        :error: ``wamp.error.invalid_argument`` WAMP error returned when ``webcluster_oid`` was invalid.
        :error: ``crossbar.error.no_such_object`` WAMP error returned when ``webcluster_oid`` was not found.
        :error: ``crossbar.error.not_stopped`` WAMP error returned when web cluster is not in status ``STOPPED``.

        :param webcluster_oid: OID of the Web cluster to delete

        :returns: Deleted router cluster, for example:

            .. code-block:: json

                {
                    "changed": 1598391866736370655,
                    "description": null,
                    "http_access_log": null,
                    "http_client_timeout": null,
                    "http_display_tracebacks": null,
                    "http_hsts": null,
                    "http_hsts_max_age": null,
                    "label": null,
                    "name": "cluster1",
                    "oid": "90d46851-0ba4-4e3d-8d9d-7a117379b212",
                    "owner_oid": "8d6e3068-900a-4fa8-a6f5-0828c8d0ee24",
                    "status": "STOPPED",
                    "tags": null,
                    "tcp_backlog": null,
                    "tcp_interface": null,
                    "tcp_port": 8080,
                    "tcp_shared": true,
                    "tcp_version": null,
                    "tls_ca_certificates": null,
                    "tls_certificate": null,
                    "tls_chain_certificates": null,
                    "tls_ciphers": null,
                    "tls_dhparam": null,
                    "tls_key": null
                }
        """
        assert details is None or isinstance(details, CallDetails)

        self.log.info('{func}.delete_webcluster(webcluster_oid={webcluster_oid}, details={details})',
                      func=hltype(self.delete_webcluster),
                      webcluster_oid=hlid(webcluster_oid),
                      details=details)

        try:
            oid = uuid.UUID(webcluster_oid)
        except Exception as e:
            raise ApplicationError('wamp.error.invalid_argument', 'invalid oid "{}"'.format(str(e)))

        if details and details.caller_authid:
            with self.gdb.begin() as txn:
                caller_oid = self.gschema.idx_users_by_email[txn, details.caller_authid]
                if not caller_oid:
                    raise ApplicationError('wamp.error.no_such_principal',
                                           'no user found for authid "{}"'.format(details.caller_authid))
        else:
            raise ApplicationError('wamp.error.no_such_principal', 'cannot map user - no caller authid available')

        with self.db.begin(write=True) as txn:
            cluster_obj = self.schema.webclusters[txn, oid]
            if cluster_obj:
                if cluster_obj.owner_oid != caller_oid:
                    raise ApplicationError('wamp.error.not_authorized',
                                           'only owner is allowed to delete router cluster')
                if cluster_obj.status != cluster.STATUS_STOPPED:
                    raise ApplicationError('crossbar.error.not_stopped')
                del self.schema.webclusters[txn, oid]
            else:
                raise ApplicationError('crossbar.error.no_such_object', 'no object with oid {} found'.format(oid))

        cluster_obj.changed = time_ns()
        self.log.info('WebCluster object deleted from database:\n{cluster_obj}', cluster_obj=cluster_obj)

        res_obj = cluster_obj.marshal()

        await self._session.publish('{}.on_webcluster_deleted'.format(self._prefix), res_obj, options=self._PUBOPTS)

        return res_obj

    @wamp.register(None, check_types=True)
    async def start_webcluster(self, webcluster_oid: str, details: Optional[CallDetails] = None) -> dict:
        """
        Start a web cluster.

        :param webcluster_oid: Object ID of web cluster to start.

        :return: Started web cluster, for example:

            .. code-block:: json

                {
                    "changed": 1598395797587541578,
                    "oid": "92f5f4c7-4175-4c72-a0a6-81467c343565",
                    "status": "STARTING",
                    "who": {
                        "authid": "superuser",
                        "authrole": "owner",
                        "session": 2587789023701533
                    }
                }
        """
        assert details is None or isinstance(details, CallDetails)

        self.log.info('{func}(webcluster_oid="{webcluster_oid}", details={details})',
                      func=hltype(self.start_webcluster),
                      webcluster_oid=hlid(webcluster_oid),
                      details=details)

        try:
            webcluster_oid_ = uuid.UUID(webcluster_oid)
        except Exception as e:
            raise ApplicationError('wamp.error.invalid_argument', 'invalid oid "{}"'.format(str(e)))

        if details and details.caller_authid:
            with self.gdb.begin() as txn:
                caller_oid = self.gschema.idx_users_by_email[txn, details.caller_authid]
                if not caller_oid:
                    raise ApplicationError('wamp.error.no_such_principal',
                                           'no user found for authid "{}"'.format(details.caller_authid))
        else:
            raise ApplicationError('wamp.error.no_such_principal', 'cannot map user - no caller authid available')

        with self.db.begin(write=True) as txn:
            webcluster = self.schema.webclusters[txn, webcluster_oid_]
            if not webcluster:
                raise ApplicationError('crossbar.error.no_such_object',
                                       'no webcluster with oid {} found'.format(webcluster_oid_))

            if webcluster.owner_oid != caller_oid:
                raise ApplicationError('wamp.error.not_authorized', 'only owner is allowed to start a web cluster')

            if webcluster.status not in [cluster.STATUS_STOPPED, cluster.STATUS_PAUSED]:
                emsg = 'cannot start webcluster currently in state {}'.format(
                    cluster.STATUS_BY_CODE[webcluster.status])
                raise ApplicationError('crossbar.error.cannot_start', emsg)

            webcluster.status = cluster.STATUS_STARTING
            webcluster.changed = time_ns()

            self.schema.webclusters[txn, webcluster_oid_] = webcluster

        monitor = WebClusterMonitor(self, webcluster_oid_)
        monitor.start()
        assert webcluster_oid_ not in self._monitors
        self._monitors[webcluster_oid_] = monitor

        webcluster_starting = {
            'oid': str(webcluster.oid),
            'status': cluster.STATUS_BY_CODE[webcluster.status],
            'changed': webcluster.changed,
            'who': {
                'session': details.caller if details else None,
                'authid': details.caller_authid if details else None,
                'authrole': details.caller_authrole if details else None,
            }
        }
        await self._session.publish('{}.on_webcluster_starting'.format(self._prefix),
                                    webcluster_starting,
                                    options=self._PUBOPTS)

        return webcluster_starting

    @wamp.register(None, check_types=True)
    async def stop_webcluster(self, webcluster_oid: str, details: Optional[CallDetails] = None) -> dict:
        """
        Stop a running web cluster.

        :procedure: ``crossbarfabriccenter.webcluster.stop_webcluster`` URI of WAMP procedure to call.
        :event: ``crossbarfabriccenter.webcluster.on_webcluster_stoppping`` WAMP event published once the web cluster is stopping.
        :error: ``wamp.error.invalid_argument`` WAMP error returned when ``webcluster_oid`` was invalid.
        :error: ``crossbar.error.no_such_object`` WAMP error returned when ``webcluster_oid`` was not found.
        :error: ``crossbar.error.cannot_stop`` WAMP error returned when web cluster is not in status ``RUNNING`` or ``STARTING``.

        :param webcluster_oid: Object ID of web cluster to stop.

        :return: Stopped web cluster, for example:

            .. code-block:: json

                {
                    "changed": 1598395878649535271,
                    "oid": "92f5f4c7-4175-4c72-a0a6-81467c343565",
                    "status": "STOPPING",
                    "who": {
                        "authid": "superuser",
                        "authrole": "owner",
                        "session": 1047738189758144
                    }
                }
        """
        assert details is None or isinstance(details, CallDetails)

        self.log.info('{func}(webcluster_oid={webcluster_oid}, details={details})',
                      webcluster_oid=hlid(webcluster_oid),
                      func=hltype(self.stop_webcluster),
                      details=details)

        try:
            webcluster_oid_ = uuid.UUID(webcluster_oid)
        except Exception as e:
            raise ApplicationError('wamp.error.invalid_argument', 'invalid oid "{}"'.format(str(e)))

        with self.db.begin(write=True) as txn:
            webcluster = self.schema.webclusters[txn, webcluster_oid_]
            if not webcluster:
                raise ApplicationError('crossbar.error.no_such_object',
                                       'no webcluster with oid {} found'.format(webcluster_oid_))

            if webcluster.status not in [cluster.STATUS_STARTING, cluster.STATUS_RUNNING]:
                emsg = 'cannot stop webcluster currently in state {}'.format(cluster.STATUS_BY_CODE[webcluster.status])
                raise ApplicationError('crossbar.error.cannot_stop', emsg)

            webcluster.status = cluster.STATUS_STOPPING
            webcluster.changed = time_ns()

            self.schema.webclusters[txn, webcluster_oid_] = webcluster

        webcluster_stopping = {
            'oid': str(webcluster.oid),
            'status': cluster.STATUS_BY_CODE[webcluster.status],
            'changed': webcluster.changed,
            'who': {
                'session': details.caller if details else None,
                'authid': details.caller_authid if details else None,
                'authrole': details.caller_authrole if details else None,
            }
        }

        await self._session.publish('{}.on_webcluster_stopping'.format(self._prefix),
                                    webcluster_stopping,
                                    options=self._PUBOPTS)

        return webcluster_stopping

    @wamp.register(None, check_types=True)
    def list_webcluster_nodes(self,
                              webcluster_oid: str,
                              return_names: Optional[bool] = None,
                              filter_by_status: Optional[str] = None,
                              details: Optional[CallDetails] = None) -> List[str]:
        """
        List nodes currently associated with the given web cluster.

        :param webcluster_oid: The web cluster to list nodes for.
        :param return_names: Return web cluster node names (WAMP authid) instead of object IDs.
        :param filter_by_status: Filter nodes by this status, eg. ``"online"``.

        :return: List of node IDs of nodes associated with the web cluster, for example:

            .. code-block:: json

                [
                    "3c7584f7-b8db-4e9b-8508-7ab4d573265d",
                    "aa1d67de-d434-4bee-96ec-d0f576c12e02",
                    "b0b36b60-5712-40fd-8ae5-ac1177ea850c",
                    "cf0241bc-b9d9-4b81-a496-5873a74b5f0a"
                ]

            or with ``return_names`` set:

            .. code-block:: json

                [
                    "node1",
                    "node2",
                    "node3",
                    "node4"
                ]
        """
        assert type(webcluster_oid) == str
        assert return_names is None or type(return_names) == bool
        assert filter_by_status is None or type(filter_by_status) == str
        assert details is None or isinstance(details, CallDetails)

        self.log.info(
            '{func}(webcluster_oid={webcluster_oid}, return_names={return_names}, filter_by_status={filter_by_status}, details={details})',
            func=hltype(self.list_webcluster_nodes),
            webcluster_oid=hlid(webcluster_oid),
            return_names=hlval(return_names),
            filter_by_status=hlval(filter_by_status),
            details=details)

        try:
            webcluster_oid_ = uuid.UUID(webcluster_oid)
        except Exception as e:
            raise ApplicationError('wamp.error.invalid_argument', 'invalid oid "{}"'.format(str(e)))

        node_oids = []
        with self.db.begin() as txn:
            webcluster = self.schema.webclusters[txn, webcluster_oid_]
            if not webcluster:
                raise ApplicationError('crossbar.error.no_such_object',
                                       'no webcluster with oid {} found'.format(webcluster_oid_))
            for _, node_oid in self.schema.webcluster_node_memberships.select(txn,
                                                                              from_key=(webcluster_oid_,
                                                                                        uuid.UUID(bytes=b'\0' * 16)),
                                                                              return_values=False):
                node_oids.append(node_oid)

        if filter_by_status:
            node_oids_ = []
            for node_oid in node_oids:
                node = self._session.nodes.get(str(node_oid), None)
                if node and node.status == filter_by_status:
                    node_oids_.append(node_oid)
            node_oids = node_oids_

        if return_names:
            node_authids = []
            with self.gdb.begin() as txn:
                for node_oid in node_oids:
                    node = self.gschema.nodes[txn, node_oid]
                    if node and node.authid:
                        node_authids.append(node.authid)
            res = sorted(node_authids)
        else:
            res = [str(node_oid) for node_oid in node_oids]

        return res

    @wamp.register(None, check_types=True)
    async def add_webcluster_node(self,
                                  webcluster_oid: str,
                                  node_oid: str,
                                  config: Optional[dict] = None,
                                  details: Optional[CallDetails] = None) -> dict:
        """
        Add a node to a web cluster. The node-to-webcluster association can be configured in ``config``:

        .. code-block:: json

            {
                "parallel": 2,
                "standby": false
            }

        * ``parallel``: The parallelism (in CPU) that the node should receive.
        * ``standby``: Wheather to add this node as a standby node that only takes over work when active nodes fail.

        The web cluster will start and manage ``parallel`` proxy workers on the node.
        When ``standby`` is set (*NOT YET IMPLEMENTED*), the node will not become active
        immediately, but only be used and become active to replace a failed (active) node
        in the cluster.

        :param webcluster_oid: OID of the web cluster to which to add the node.
        :param node_oid: OID of the node to add to the cluster. A node can be added to more than one cluster.

        :return: Added node, for example:

            .. code-block:: json

                {
                    "cluster_oid": "92f5f4c7-4175-4c72-a0a6-81467c343565",
                    "node_oid": "b0b36b60-5712-40fd-8ae5-ac1177ea850c",
                    "parallel": 2,
                    "standby": null
                }
        """
        assert details is None or isinstance(details, CallDetails)

        self.log.info(
            '{func}(webcluster_oid={webcluster_oid}, node_oid={node_oid}, config={config}, details={details})',
            func=hltype(self.list_webcluster_nodes),
            webcluster_oid=hlid(webcluster_oid),
            node_oid=hlid(node_oid),
            config=config,
            details=details)

        config = config or {}
        config['cluster_oid'] = webcluster_oid
        config['node_oid'] = node_oid
        membership = WebClusterNodeMembership.parse(config)

        with self.gdb.begin() as txn:
            node = self.gschema.nodes[txn, membership.node_oid]
            if not node or node.mrealm_oid != self._session._mrealm_oid:
                raise ApplicationError('crossbar.error.no_such_object',
                                       'no node with oid {} found'.format(membership.node_oid))

        with self.db.begin(write=True) as txn:
            webcluster = self.schema.webclusters[txn, membership.cluster_oid]
            if not webcluster:
                raise ApplicationError('crossbar.error.no_such_object',
                                       'no webcluster with oid {} found'.format(membership.cluster_oid))

            self.schema.webcluster_node_memberships[txn, (membership.cluster_oid, membership.node_oid)] = membership

        res_obj = membership.marshal()
        self.log.info('node added to web cluster:\n{membership}', membership=res_obj)

        await self._session.publish('{}.on_webcluster_node_added'.format(self._prefix), res_obj, options=self._PUBOPTS)

        return res_obj

    @wamp.register(None, check_types=True)
    async def remove_webcluster_node(self,
                                     webcluster_oid: str,
                                     node_oid: str,
                                     details: Optional[CallDetails] = None) -> dict:
        """
        Remove a node from a web cluster.

        :param webcluster_oid: OID of the web cluster from which to remove the node.
        :param node_oid: OID of the node to remove from the web cluster

        :return: Node removed from web cluster, for example:

            .. code-block:: json

                {
                    "cluster_oid": "92f5f4c7-4175-4c72-a0a6-81467c343565",
                    "node_oid": "cf0241bc-b9d9-4b81-a496-5873a74b5f0a",
                    "parallel": 2,
                    "standby": null
                }
        """
        assert type(webcluster_oid) == str
        assert type(node_oid) == str
        assert details is None or isinstance(details, CallDetails)

        try:
            node_oid_ = uuid.UUID(node_oid)
        except Exception as e:
            raise ApplicationError('wamp.error.invalid_argument', 'invalid oid "{}"'.format(str(e)))

        try:
            webcluster_oid_ = uuid.UUID(webcluster_oid)
        except Exception as e:
            raise ApplicationError('wamp.error.invalid_argument', 'invalid oid "{}"'.format(str(e)))

        self.log.info('{func}(webcluster_oid={webcluster_oid}, node_oid={node_oid}, details={details})',
                      webcluster_oid=hlid(webcluster_oid_),
                      node_oid=hlid(node_oid_),
                      func=hltype(self.remove_webcluster_node),
                      details=details)

        with self.gdb.begin() as txn:
            node = self.gschema.nodes[txn, node_oid_]
            if not node or node.mrealm_oid != self._session._mrealm_oid:
                raise ApplicationError('crossbar.error.no_such_object', 'no node with oid {} found'.format(node_oid_))

        with self.db.begin(write=True) as txn:
            webcluster = self.schema.webclusters[txn, webcluster_oid_]
            if not webcluster:
                raise ApplicationError('crossbar.error.no_such_object',
                                       'no object with oid {} found'.format(webcluster_oid_))

            membership = self.schema.webcluster_node_memberships[txn, (webcluster_oid_, node_oid_)]
            if not membership:
                raise ApplicationError(
                    'crossbar.error.no_such_object',
                    'no association between node {} and webcluster {} found'.format(node_oid_, webcluster_oid_))

            del self.schema.webcluster_node_memberships[txn, (webcluster_oid_, node_oid_)]

        res_obj = membership.marshal()
        self.log.info('node removed from web cluster:\n{res_obj}', membership=res_obj)

        await self._session.publish('{}.on_webcluster_node_removed'.format(self._prefix),
                                    res_obj,
                                    options=self._PUBOPTS)

        return res_obj

    @wamp.register(None, check_types=True)
    def get_webcluster_node(self, webcluster_oid: str, node_oid: str, details: Optional[CallDetails] = None) -> dict:
        """
        Get information (such as for example parallel degree) for the association
        of a node with a web cluster.

        :param webcluster_oid: Object ID of web cluster to return node association for.
        :param node_oid: Object ID of node to return association for.

        :return: Information for the association of the node with the webcluster, for example:

            .. code-block:: json

                {
                    "cluster_oid": "92f5f4c7-4175-4c72-a0a6-81467c343565",
                    "node_oid": "cf0241bc-b9d9-4b81-a496-5873a74b5f0a",
                    "parallel": 2,
                    "standby": null
                }
        """
        assert type(webcluster_oid) == str
        assert type(node_oid) == str
        assert details is None or isinstance(details, CallDetails)

        try:
            node_oid_ = uuid.UUID(node_oid)
        except Exception as e:
            raise ApplicationError('wamp.error.invalid_argument', 'invalid oid "{}"'.format(str(e)))

        try:
            webcluster_oid_ = uuid.UUID(webcluster_oid)
        except Exception as e:
            raise ApplicationError('wamp.error.invalid_argument', 'invalid oid "{}"'.format(str(e)))

        self.log.info('{func}(webcluster_oid={webcluster_oid}, node_oid={node_oid}, details={details})',
                      webcluster_oid=hlid(webcluster_oid_),
                      node_oid=hlid(node_oid_),
                      func=hltype(self.get_webcluster_node),
                      details=details)

        with self.gdb.begin() as txn:
            node = self.gschema.nodes[txn, node_oid_]
            if not node or node.mrealm_oid != self._session._mrealm_oid:
                raise ApplicationError('crossbar.error.no_such_object', 'no node with oid {} found'.format(node_oid_))

        with self.db.begin() as txn:
            webcluster = self.schema.webclusters[txn, webcluster_oid_]
            if not webcluster:
                raise ApplicationError('crossbar.error.no_such_object',
                                       'no object with oid {} found'.format(webcluster_oid_))

            membership = self.schema.webcluster_node_memberships[txn, (webcluster_oid_, node_oid_)]
            if not membership:
                raise ApplicationError(
                    'crossbar.error.no_such_object',
                    'no association between node {} and webcluster {} found'.format(node_oid_, webcluster_oid_))

        res_obj = membership.marshal()

        return res_obj

    @wamp.register(None, check_types=True)
    def list_webcluster_services(self,
                                 webcluster_oid: str,
                                 prefix: Optional[str] = None,
                                 details: Optional[CallDetails] = None) -> Dict[str, str]:
        """
        List webservices defined on a webcluster, optionally filtering by prefix.

        :param webcluster_oid: The web cluster for which to list currently defined web services.
        :param prefix: If provided, the path prefix for filtering web services.

        :return: A map with HTTP paths as keys and webservice ID as values, for example:

            .. code-block:: json

                {
                    "/": "92f5f4c7-4175-4c72-a0a6-81467c343565",
                    "info": "92f5f4c7-4175-4c72-a0a6-81467c343565",
                    "settings": "92f5f4c7-4175-4c72-a0a6-81467c343565",
                    "ws": "92f5f4c7-4175-4c72-a0a6-81467c343565"
                }
        """
        assert type(webcluster_oid) == str
        assert prefix is None or type(prefix) == str
        assert details is None or isinstance(details, CallDetails)

        self.log.info('{func}(webcluster_oid={webcluster_oid}, prefix="{prefix}", details={details})',
                      webcluster_oid=hlid(webcluster_oid),
                      prefix=hlval(prefix),
                      func=hltype(self.list_webcluster_services),
                      details=details)

        if prefix:
            raise NotImplementedError('prefix option not yet implemented')

        try:
            webcluster_oid_ = uuid.UUID(webcluster_oid)
        except Exception as e:
            raise ApplicationError('wamp.error.invalid_argument', 'invalid oid "{}"'.format(str(e)))

        with self.db.begin() as txn:
            webcluster = self.schema.webclusters[txn, webcluster_oid_]
            if not webcluster:
                raise ApplicationError('crossbar.error.no_such_object',
                                       'no object with oid {} found'.format(webcluster_oid_))

            res = {}
            # FIXME: 1) to_key and 2) prefix option
            for (_, path), webservice_oid in self.schema.idx_webservices_by_path.select(txn,
                                                                                        from_key=(webcluster_oid_,
                                                                                                  '')):
                res[path] = str(webservice_oid)

            return res

    @wamp.register(None, check_types=True)
    async def add_webcluster_service(self,
                                     webcluster_oid: str,
                                     path: str,
                                     webservice: dict,
                                     details: Optional[CallDetails] = None) -> dict:
        """
        Add a Web service to a Web cluster.

        :param webcluster_oid: Web cluster to which to add the Web service.
        :param path: The path on which to add the webservice.
        :param webservice: Web service definition object.

        :returns: The web service added to the web cluster, for example:

            .. code-block:: json

                {
                    "description": null,
                    "directory": "..",
                    "label": null,
                    "oid": "4411eb80-006a-45a0-8624-52a7ee84d0ce",
                    "options": {
                        "enable_directory_listing": true
                    },
                    "path": "/",
                    "tags": null,
                    "type": "static",
                    "webcluster_oid": "92f5f4c7-4175-4c72-a0a6-81467c343565"
                }
        """
        assert type(webcluster_oid) == str
        assert type(path) == str
        assert type(webservice) == dict
        assert details is None or isinstance(details, CallDetails)

        self.log.info(
            '{func}(webcluster_oid={webcluster_oid}, path={path}, webservice={webservice}, details={details})',
            webcluster_oid=hlid(webcluster_oid),
            path=hlval(path),
            webservice=webservice,
            func=hltype(self.add_webcluster_service),
            details=details)

        try:
            webcluster_oid_ = uuid.UUID(webcluster_oid)
        except Exception as e:
            raise ApplicationError('wamp.error.invalid_argument', 'invalid oid "{}"'.format(str(e)))

        if details and details.caller_authid:
            with self.gdb.begin() as txn:
                caller_oid = self.gschema.idx_users_by_email[txn, details.caller_authid]
                if not caller_oid:
                    raise ApplicationError('wamp.error.no_such_principal',
                                           'no user found for authid "{}"'.format(details.caller_authid))
        else:
            raise ApplicationError('wamp.error.no_such_principal', 'cannot map user - no caller authid available')

        if 'type' not in webservice:
            raise ApplicationError('wamp.error.invalid_argument', 'missing type in webservice')

        webservice_type = webservice['type']
        if webservice_type not in self._WEB_SERVICE_CHECKERS:
            raise ApplicationError('wamp.error.invalid_argument',
                                   'invalid webservice type "{}"'.format(webservice_type))

        if 'path' in webservice:
            assert webservice['path'] == path
            del webservice['path']

        _personality = self._worker.personality
        _check = self._WEB_SERVICE_CHECKERS[webservice_type]
        self.log.info('Checking web service configuration [personality={personality}, check={check}]:\n{webservice}',
                      personality=_personality,
                      check=_check,
                      webservice=pformat(webservice))

        try:
            _check(_personality, webservice)
        except Exception as e:
            raise ApplicationError('wamp.error.invalid_argument',
                                   'invalid webservice configuration for type "{}": {}'.format(webservice_type, e))

        try:
            webservice_obj = WebService.parse(webservice)
        except Exception as e:
            raise ApplicationError('wamp.error.invalid_configuration',
                                   'could not parse web cluster configuration ({})'.format(e))

        webservice_obj.oid = uuid.uuid4()
        webservice_obj.path = path
        webservice_obj.webcluster_oid = webcluster_oid_

        with self.db.begin(write=True) as txn:
            webcluster = self.schema.webclusters[txn, webservice_obj.webcluster_oid]
            if not webcluster:
                raise ApplicationError('crossbar.error.no_such_object',
                                       'no object with oid {} found'.format(webservice_obj.webcluster_oid))

            if webcluster.owner_oid != caller_oid:
                raise ApplicationError('wamp.error.not_authorized',
                                       'only owner is allowed to add web service to a web cluster')

            if webcluster.status not in [cluster.STATUS_STOPPED, cluster.STATUS_PAUSED]:
                emsg = 'cannot add web service to webcluster currently in state {}'.format(
                    cluster.STATUS_BY_CODE[webcluster.status])
                raise ApplicationError('crossbar.error.cannot_start', emsg)

            self.schema.webservices[txn, webservice_obj.oid] = webservice_obj

            webcluster.changed = time_ns()
            self.schema.webclusters[txn, webcluster.oid] = webcluster

        self.log.info('New WebService object stored in database:\n{webservice_obj}', webservice_obj=webservice_obj)

        res_obj = webservice_obj.marshal()

        await self._session.publish('{}.on_webservice_added'.format(self._prefix), res_obj, options=self._PUBOPTS)

        self.log.info('Management API event <on_webservice_added> published:\n{res_obj}', res_obj=res_obj)

        return res_obj

    @wamp.register(None, check_types=True)
    async def remove_webcluster_service(self,
                                        webcluster_oid: str,
                                        webservice_oid: str,
                                        details: Optional[CallDetails] = None) -> dict:
        """
        Remove the Web service from the Web cluster.

        :param webcluster_oid: Object ID of the web cluster from which to remove the web service.
        :param webservice_oid: Object ID of the web service to remove.

        :return: The web service removed from the web cluster, for example:

            .. code-block:: json

                {
                    "description": null,
                    "label": null,
                    "oid": "96d30fa8-6df2-4888-9845-87db8884a062",
                    "path": "settings",
                    "tags": null,
                    "type": "json",
                    "value": [
                        1,
                        2,
                        3
                    ],
                    "webcluster_oid": "92f5f4c7-4175-4c72-a0a6-81467c343565"
                }
        """
        assert type(webcluster_oid) == str
        assert type(webservice_oid) == str
        assert details is None or isinstance(details, CallDetails)

        self.log.info(
            '{klass}.remove_webcluster_service(webcluster_oid={webcluster_oid}, webservice_oid={webservice_oid}, details={details})',
            webcluster_oid=hlid(webcluster_oid),
            webservice_oid=hlid(webservice_oid),
            func=hltype(self.remove_webcluster_service),
            details=details)

        try:
            webcluster_oid_ = uuid.UUID(webcluster_oid)
        except Exception as e:
            raise ApplicationError('wamp.error.invalid_argument', 'invalid oid "{}"'.format(str(e)))

        try:
            webservice_oid_ = uuid.UUID(webservice_oid)
        except Exception as e:
            raise ApplicationError('wamp.error.invalid_argument', 'invalid oid "{}"'.format(str(e)))

        if details and details.caller_authid:
            with self.gdb.begin() as txn:
                caller_oid = self.gschema.idx_users_by_email[txn, details.caller_authid]
                if not caller_oid:
                    raise ApplicationError('wamp.error.no_such_principal',
                                           'no user found for authid "{}"'.format(details.caller_authid))
        else:
            raise ApplicationError('wamp.error.no_such_principal', 'cannot map user - no caller authid available')

        with self.db.begin(write=True) as txn:
            webcluster = self.schema.webclusters[txn, webcluster_oid_]
            if not webcluster:
                raise ApplicationError('crossbar.error.no_such_object',
                                       'no object with oid {} found'.format(webcluster_oid_))

            if webcluster.owner_oid != caller_oid:
                raise ApplicationError('wamp.error.not_authorized',
                                       'only owner is allowed to remove a web service from a web cluster')

            webservice = self.schema.webservices[txn, webservice_oid_]
            if not webservice:
                raise ApplicationError('crossbar.error.no_such_object',
                                       'no object with oid {} found'.format(webservice_oid_))

            del self.schema.webservices[txn, webservice_oid_]

        self.log.info('WebService object removed from database:\n{webservice}', webservice=webservice)

        res_obj = webservice.marshal()

        await self._session.publish('{}.on_webservice_removed'.format(self._prefix), res_obj, options=self._PUBOPTS)

        self.log.info('Management API event <on_webservice_removed> published:\n{res_obj}', res_obj=res_obj)

        return res_obj

    @wamp.register(None, check_types=True)
    def get_webcluster_service(self,
                               webcluster_oid: str,
                               webservice_oid: str,
                               details: Optional[CallDetails] = None) -> dict:
        """
        Get definition of a web service by ID.

        See also the corresponding procedure :meth:`crossbar.master.cluster.webcluster.WebClusterManager.get_webcluster_service_by_path`
        which returns the same information, given HTTP path rather than object ID.

        :param webcluster_oid: The web cluster running the web service to return.
        :param webservice_oid: The web service to return.

        :return: The web service definition, for example:

            .. code-block:: json

                {
                    "description": null,
                    "label": null,
                    "oid": "6cc51192-4259-4640-84cb-ee03b6f92fbf",
                    "path": "info",
                    "tags": null,
                    "type": "nodeinfo",
                    "webcluster_oid": "92f5f4c7-4175-4c72-a0a6-81467c343565"
                }
        """
        assert type(webcluster_oid) == str
        assert type(webservice_oid) == str
        assert details is None or isinstance(details, CallDetails)

        self.log.info('{func}(webcluster_oid={webcluster_oid}, webservice_oid={webservice_oid}, details={details})',
                      webcluster_oid=hlid(webcluster_oid),
                      webservice_oid=hlid(webservice_oid),
                      func=hltype(self.get_webcluster_service),
                      details=details)

        try:
            webcluster_oid_ = uuid.UUID(webcluster_oid)
        except Exception as e:
            raise ApplicationError('wamp.error.invalid_argument', 'invalid oid "{}"'.format(str(e)))

        try:
            webservice_oid_ = uuid.UUID(webservice_oid)
        except Exception as e:
            raise ApplicationError('wamp.error.invalid_argument', 'invalid oid "{}"'.format(str(e)))

        with self.db.begin() as txn:
            webcluster = self.schema.webclusters[txn, webcluster_oid_]
            if not webcluster:
                raise ApplicationError('crossbar.error.no_such_object',
                                       'no object with oid {} found'.format(webcluster_oid_))

            webservice = self.schema.webservices[txn, webservice_oid_]
            if not webservice:
                raise ApplicationError('crossbar.error.no_such_object',
                                       'no object with oid {} found'.format(webservice_oid_))

        return webservice.marshal()

    @wamp.register(None, check_types=True)
    def get_webcluster_service_by_path(self,
                                       webcluster_oid: str,
                                       path: str,
                                       details: Optional[CallDetails] = None) -> dict:
        """
        Get definition of a web service by HTTP path.

        See also the corresponding procedure :meth:`crossbar.master.cluster.webcluster.WebClusterManager.get_webcluster_service`
        which returns the same information, given and object ID rather than HTTP path.

        :param webcluster_oid: The web cluster running the web service to return.
        :param path: HTTP path of web service to return.

        :return: see :meth:`crossbar.master.cluster.webcluster.WebClusterManager.get_webcluster_service`
        """
        assert type(webcluster_oid) == str
        assert type(path) == str
        assert details is None or isinstance(details, CallDetails)

        self.log.info('{func}(webcluster_oid={webcluster_oid}, path="{path}", details={details})',
                      webcluster_oid=hlid(webcluster_oid),
                      path=hlval(path),
                      func=hltype(self.get_webcluster_service_by_path),
                      details=details)

        try:
            webcluster_oid_ = uuid.UUID(webcluster_oid)
        except Exception as e:
            raise ApplicationError('wamp.error.invalid_argument', 'invalid oid "{}"'.format(str(e)))

        with self.db.begin() as txn:
            webcluster = self.schema.webclusters[txn, webcluster_oid_]
            if not webcluster:
                raise ApplicationError('crossbar.error.no_such_object',
                                       'no object with oid {} found'.format(webcluster_oid_))

            webservice_oid = self.schema.idx_webservices_by_path[txn, (webcluster_oid_, path)]

            if not webservice_oid:
                raise ApplicationError('crossbar.error.no_such_object',
                                       'no webservice for path "{}" found'.format(path))

            webservice = self.schema.webservices[txn, webservice_oid]
            assert webservice

        return webservice.marshal()



###############################################################################
#
# Crossbar.io Master
# Copyright (c) Crossbar.io Technologies GmbH. Licensed under EUPLv1.2.
#
###############################################################################

from typing import List, Union  # noqa

from crossbar.master.api.remote import RemoteApi  # noqa
from crossbar.master.api.node import RemoteNodeApi
# from crossbar.master.api.nativeprocess import RemoteNativeProcessApi
from crossbar.master.api.worker import RemoteWorkerApi
from crossbar.master.api.router import RemoteRouterApi
from crossbar.master.api.container import RemoteContainerApi
from crossbar.master.api.proxy import RemoteProxyApi
from crossbar.master.api.tracing import RemoteTracingApi
from crossbar.master.api.docker import RemoteDockerApi
from crossbar.master.api.wamp import RemoteWampApi

APIS = []  # type: List[Union[RemoteApi, RemoteWampApi]]
APIS.append(RemoteNodeApi())
# APIS.append(RemoteNativeProcessApi())
APIS.append(RemoteWorkerApi())
APIS.append(RemoteRouterApi())
APIS.append(RemoteContainerApi())
APIS.append(RemoteProxyApi())
APIS.append(RemoteTracingApi())
APIS.append(RemoteDockerApi())
APIS.append(RemoteWampApi())



#####################################################################################
#
#  Copyright (c) Crossbar.io Technologies GmbH
#  SPDX-License-Identifier: EUPL-1.2
#
#####################################################################################

import sys
import importlib

from twisted.python.threadpool import ThreadPool
from twisted.web.wsgi import WSGIResource

from autobahn.wamp import ApplicationError
from autobahn.twisted.resource import WSGIRootResource

from crossbar.webservice.base import RouterWebService


class RouterWebServiceWsgi(RouterWebService):
    """
    WSGI application Web service.
    """
    @staticmethod
    def create(transport, path, config):
        personality = transport.worker.personality
        personality.WEB_SERVICE_CHECKERS['wsgi'](personality, config)
        reactor = transport.worker.components_shared['reactor']

        if 'module' not in config:
            raise ApplicationError('crossbar.error.invalid_configuration', 'missing WSGI app module')

        if 'object' not in config:
            raise ApplicationError('crossbar.error.invalid_configuration', 'missing WSGI app object')

        # import WSGI app module and object
        mod_name = config['module']
        try:
            mod = importlib.import_module(mod_name)
        except ImportError as e:
            raise ApplicationError(
                'crossbar.error.invalid_configuration',
                'WSGI app module "{}" import failed: {} - Python search path was {}'.format(mod_name, e, sys.path))

        obj_name = config['object']
        if obj_name not in mod.__dict__:
            raise ApplicationError('crossbar.error.invalid_configuration',
                                   'WSGI app object "{}" not in module "{}"'.format(obj_name, mod_name))
        else:
            app = getattr(mod, obj_name)

        # Create a thread-pool for running the WSGI requests in
        pool = ThreadPool(maxthreads=config.get('maxthreads', 20),
                          minthreads=config.get('minthreads', 0),
                          name='crossbar_wsgi_threadpool')
        reactor.addSystemEventTrigger('before', 'shutdown', pool.stop)
        pool.start()

        # Create a Twisted Web WSGI resource from the user's WSGI application object
        try:
            resource = WSGIResource(reactor, pool, app)
            if path == '/':
                resource = WSGIRootResource(resource, {})
        except Exception as e:
            raise ApplicationError('crossbar.error.invalid_configuration',
                                   'could not instantiate WSGI resource: {}'.format(e))
        else:
            return RouterWebServiceWsgi(transport, path, config, resource)



##############################################################################
#
#                        Crossbar.io
#     Copyright (C) Crossbar.io Technologies GmbH. All rights reserved.
#
##############################################################################

import os
import time
import threading

import six
import psutil
from twisted.internet.defer import inlineCallbacks, DeferredList
from twisted.internet.threads import deferToThread

from txaio import make_logger, time_ns
from autobahn import wamp
from autobahn.wamp.types import PublishOptions
from autobahn.wamp.exception import ApplicationError
from crossbar.node.worker import NativeWorkerProcess
from crossbar.worker.controller import WorkerController

from crossbar.edge.worker.monitor import MONITORS

__all__ = ('HostMonitor', 'HostMonitorProcess')


class HostMonitorProcess(NativeWorkerProcess):

    TYPE = 'hostmonitor'
    LOGNAME = 'HostMon'


class HostMonitor(WorkerController):
    """
    Gather resource usage statistics and feed them back to the client
    """

    WORKER_TYPE = u'hostmonitor'
    WORKER_TITLE = u'HostMonitor'

    log = make_logger()

    def __init__(self, config=None, reactor=None, personality=None):
        super(HostMonitor, self).__init__(config=config, reactor=reactor, personality=personality)

        self._process = psutil.Process(os.getpid())

        self._run = False
        self._is_running = False
        self._monitors = {}
        self._config = None

    @inlineCallbacks
    def onJoin(self, details):
        self.log.info('HostMonitor connected (monitors available: {monitors})', monitors=sorted(MONITORS.keys()))

        yield WorkerController.onJoin(self, details, publish_ready=False)

        # register monitor procedures
        dl = []
        for monitor in self._monitors.values():
            d = self.register(monitor.get, u'{}.get_{}'.format(self._prefix, monitor.ID))
            dl.append(d)
        res = yield DeferredList(dl, fireOnOneErrback=True)
        print(res)
        self.log.info('HostMonitor {pcnt} procedures registered', pcnt=len(res))

        # signal this worker is done with setup and ready
        yield self.publish_ready()

    def onLeave(self, details):
        self.log.info('HostMonitor shutting down ..')
        self.stop_monitoring()
        self.disconnect()

    @wamp.register(None)
    def get_monitoring(self, sensors=None, details=None):
        if self._run:
            if self._is_running:
                status = u'running'
            else:
                status = u'starting'
            pass
        else:
            status = u'stopped'

        if sensors is None:
            sensors = list(self._monitors.keys())
        else:
            sensors = list(set(self._monitors.keys()).intersection(sensors))

        monitoring = {
            u'status': status,
            u'config': self._config,
            u'current': {sensor: self._monitors[sensor].get()
                         for sensor in sensors}
        }
        return monitoring

    @wamp.register(None)
    def start_monitoring(self, config, details=None):
        """
        Start monitoring.

        :param config: Monitoring configuration.
        :type config: dict
        """
        self.log.info('HostMonitor start monitoring (config={config})', config=config)

        # after the worker is up and running, host monitoring must be started
        # (host monitoring can only be started or off, there is nothing to start instances of)
        if self._run:
            if self._is_running:
                raise ApplicationError(u'crossbar.error.already_running',
                                       u'cannot start host monitoring - monitoring already running')
            else:
                raise ApplicationError(u'crossbar.error.already_running',
                                       u'cannot start host monitoring - monitoring currently starting')

        # submonitors specified in the config
        monitors = config.get('monitors', {})
        if len(monitors) == 0:
            raise ApplicationError(u'crossbar.error.invalid_configuration',
                                   u'cannot start monitoring from empty monitors list')

        # save config for later access
        self._config = config

        # sensor polling interval in ms
        self._interval = config.get('interval', 500)

        self._monitors = {}
        for monitor_key, monitor_config in monitors.items():

            if type(monitor_key) != six.text_type:
                raise ApplicationError(u'crossbar.error.invalid_configuration',
                                       u'invalid monitor key type "{}"'.format(type(monitor_key)))

            if monitor_key not in MONITORS.keys():
                raise ApplicationError(
                    u'crossbar.error.invalid_configuration',
                    u'unknown monitor type "{}" (available monitors: {})'.format(monitor_key, sorted(MONITORS.keys())))

            # get submonitor class
            klass = MONITORS[monitor_key]

            # instantiate submonitor
            self._monitors[monitor_key] = klass(monitor_config)

        def after_exit_success(_):
            self._run = False

        def after_exit_error(err):
            self._run = False

        self._run = True
        d = deferToThread(self._loop)
        d.addCallbacks(after_exit_success, after_exit_error)

        self.log.info('HostMonitor loop started _from_ main thread (PID={pid}, thread={tid})',
                      pid=os.getpid(),
                      tid=threading.get_ident())

        started = {u'monitors': sorted(list(self._monitors.keys()))}
        topic = u'{}.on_mon_started'.format(self._uri_prefix)

        self.publish(topic, started)

        self.log.info('HostMonitor started monitoring (started={started})', started=started)

        return started

    @wamp.register(None)
    def stop_monitoring(self, details=None):
        self.log.info('HostMonitor stop monitoring ..')

        if not self._run:
            if self._is_running:
                self.log.warn('cannot stop host monitoring - monitoring already told to stop (but still running)')
            else:
                self.log.info('cannot stop host monitoring - monitoring already stopped')
            return None

        stopped = {u'monitors': sorted(list(self._monitors.keys()))}
        topic = u'{}.on_monitoring_stopped'.format(self._uri_prefix)

        self._run = False
        self._monitors = {}
        self._config = None

        self.publish(topic, stopped)

        self.log.info('HostMonitor stopped monitoring (stopped={stopped})', stopped=stopped)

        return stopped

    def _loop(self):
        # this is running on a background thread! eg you cannot just call self.publish() or self.log()!

        self._is_running = True

        try:
            print('HostMonitor entering loop on background thread (PID={pid}, thread={tid})'.format(
                pid=os.getpid(), tid=threading.get_ident()))
            while self._run:
                started = time_ns()

                self.log.debug(
                    'HostMonitor gather sensor data on background thread (started={started}, PID={pid}=thread {tid})',
                    started=started,
                    pid=os.getpid(),
                    tid=threading.get_ident())

                # poll all configured submonitors and collect the data
                hdata = {}
                for monitor in self._monitors.values():
                    hdata[monitor.ID] = monitor.poll()

                self._reactor.callFromThread(self._publish, hdata)

                # next time we want to loop (takes into account time for monitoring)
                next_time = started + self._interval * 10**6

                # wake up every 100ms until we want to loop
                while self._run and time_ns() < next_time:
                    time.sleep(.1)
        except Exception as e:
            print('HostMonitor ending loop because of exception: {}'.format(e))

            # mark the thread as "done" when still on the background thread
            self._is_running = False

            # the deferred return on the main thread from deferToThread will fire its errback
            raise

        else:
            print('HostMonitor ending loop gracefully')
            # the deferred return on the main thread from deferToThread will fire its callback

    def _publish(self, hdata):
        self.log.debug('HostMonitor publish sensor data on main thread (PID {pid} thread {tid})',
                       pid=os.getpid(),
                       tid=threading.get_ident())

        # this is running on the main thread: doing the publishes here avoids a couple
        # of context switches, and this way the background thread doesn't touch any
        # WAMP stuff (which is good in general .. decoupling)
        options = PublishOptions(acknowledge=True)

        dl = []
        for monitor_id, monitor_data in hdata.items():
            d = self.publish(u'{}.on_{}_sample'.format(self._uri_prefix, monitor_id), monitor_data, options=options)
            dl.append(d)

        d = DeferredList(dl)

        def done(results):
            ok = 0
            err = 0
            for success, result in results:
                if success:
                    ok += 1
                else:
                    err += 1
                    self.log.warn('HostMonitor publication failed with - {error}', error=result)
            self.log.debug('HostMonitor publish: ok={ok}, err={err}', ok=ok, err=err)

        d.addCallback(done)



##############################################################################
#
#                        Crossbar.io
#     Copyright (C) Crossbar.io Technologies GmbH. All rights reserved.
#
##############################################################################

import txaio

txaio.use_twisted()  # noqa

from autobahn import xbr
from crossbar._version import __version__, __build__
from crossbar.edge.personality import Personality

__all__ = ('__version__', '__build__', 'Personality', 'xbr')
__doc__ = Personality.DESC
